from typing import Any
from typing import Optional

from sqlalchemy import literal_column as literal_column
from sqlalchemy import text as text
from sqlalchemy.sql import visitors as visitors
from .enumerated import ENUM as ENUM
from .enumerated import SET as SET
from .json import JSON as JSON
from .json import JSONIndexType as JSONIndexType
from .json import JSONPathType as JSONPathType
from .types import BIGINT as BIGINT
from .types import BIT as BIT
from .types import CHAR as CHAR
from .types import DATETIME as DATETIME
from .types import DECIMAL as DECIMAL
from .types import DOUBLE as DOUBLE
from .types import FLOAT as FLOAT
from .types import INTEGER as INTEGER
from .types import LONGBLOB as LONGBLOB
from .types import LONGTEXT as LONGTEXT
from .types import MEDIUMBLOB as MEDIUMBLOB
from .types import MEDIUMINT as MEDIUMINT
from .types import MEDIUMTEXT as MEDIUMTEXT
from .types import NCHAR as NCHAR
from .types import NUMERIC as NUMERIC
from .types import NVARCHAR as NVARCHAR
from .types import REAL as REAL
from .types import SMALLINT as SMALLINT
from .types import TEXT as TEXT
from .types import TIME as TIME
from .types import TIMESTAMP as TIMESTAMP
from .types import TINYBLOB as TINYBLOB
from .types import TINYINT as TINYINT
from .types import TINYTEXT as TINYTEXT
from .types import VARCHAR as VARCHAR
from .types import YEAR as YEAR
from ... import exc as exc
from ... import log as log
from ... import sql as sql
from ... import util as util
from ...engine import default as default
from ...engine import reflection as reflection
from ...sql import coercions as coercions
from ...sql import compiler as compiler
from ...sql import elements as elements
from ...sql import functions as functions
from ...sql import operators as operators
from ...sql import roles as roles
from ...sql.sqltypes import Unicode as Unicode
from ...types import BINARY as BINARY
from ...types import BLOB as BLOB
from ...types import BOOLEAN as BOOLEAN
from ...types import DATE as DATE
from ...types import VARBINARY as VARBINARY
from ...util import compat as compat
from ...util import topological as topological

RESERVED_WORDS: Any
AUTOCOMMIT_RE: Any
SET_RE: Any
MSTime = TIME
MSSet = SET
MSEnum = ENUM
MSLongBlob = LONGBLOB
MSMediumBlob = MEDIUMBLOB
MSTinyBlob = TINYBLOB
MSBlob = BLOB
MSBinary = BINARY
MSVarBinary = VARBINARY
MSNChar = NCHAR
MSNVarChar = NVARCHAR
MSChar = CHAR
MSString = VARCHAR
MSLongText = LONGTEXT
MSMediumText = MEDIUMTEXT
MSTinyText = TINYTEXT
MSText = TEXT
MSYear = YEAR
MSTimeStamp = TIMESTAMP
MSBit = BIT
MSSmallInteger = SMALLINT
MSTinyInteger = TINYINT
MSMediumInteger = MEDIUMINT
MSBigInteger = BIGINT
MSNumeric = NUMERIC
MSDecimal = DECIMAL
MSDouble = DOUBLE
MSReal = REAL
MSFloat = FLOAT
MSInteger = INTEGER
colspecs: Any
ischema_names: Any

class MySQLExecutionContext(default.DefaultExecutionContext):
    def should_autocommit_text(self, statement: Any): ...
    def create_server_side_cursor(self): ...
    def fire_sequence(self, seq: Any, type_: Any): ...

class MySQLCompiler(compiler.SQLCompiler):
    render_table_with_column_in_update_from: bool = ...
    extract_map: Any = ...
    def default_from(self): ...
    def visit_random_func(self, fn: Any, **kw: Any): ...
    def visit_sequence(self, seq: Any, **kw: Any): ...
    def visit_sysdate_func(self, fn: Any, **kw: Any): ...
    def visit_json_getitem_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def visit_json_path_getitem_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def visit_on_duplicate_key_update(self, on_duplicate: Any, **kw: Any): ...
    def visit_concat_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def visit_match_op_binary(self, binary: Any, operator: Any, **kw: Any): ...
    def get_from_hint_text(self, table: Any, text: Any): ...
    def visit_typeclause(
        self, typeclause: Any, type_: Optional[Any] = ..., **kw: Any
    ): ...
    def visit_cast(self, cast: Any, **kw: Any): ...
    def render_literal_value(self, value: Any, type_: Any): ...
    def visit_true(self, element: Any, **kw: Any): ...
    def visit_false(self, element: Any, **kw: Any): ...
    def get_select_precolumns(self, select: Any, **kw: Any): ...
    def visit_join(
        self,
        join: Any,
        asfrom: bool = ...,
        from_linter: Optional[Any] = ...,
        **kwargs: Any,
    ): ...
    def for_update_clause(self, select: Any, **kw: Any): ...
    def limit_clause(self, select: Any, **kw: Any): ...
    def update_limit_clause(self, update_stmt: Any): ...
    def update_tables_clause(
        self, update_stmt: Any, from_table: Any, extra_froms: Any, **kw: Any
    ): ...
    def update_from_clause(
        self,
        update_stmt: Any,
        from_table: Any,
        extra_froms: Any,
        from_hints: Any,
        **kw: Any,
    ) -> None: ...
    def delete_table_clause(
        self, delete_stmt: Any, from_table: Any, extra_froms: Any
    ): ...
    def delete_extra_from_clause(
        self,
        delete_stmt: Any,
        from_table: Any,
        extra_froms: Any,
        from_hints: Any,
        **kw: Any,
    ): ...
    def visit_empty_set_expr(self, element_types: Any): ...
    def visit_is_distinct_from_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def visit_is_not_distinct_from_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def visit_regexp_match_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def visit_not_regexp_match_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...
    def visit_regexp_replace_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ): ...

class MySQLDDLCompiler(compiler.DDLCompiler):
    def get_column_specification(self, column: Any, **kw: Any): ...
    def post_create_table(self, table: Any): ...
    def visit_create_index(self, create: Any, **kw: Any): ...
    def visit_primary_key_constraint(self, constraint: Any): ...
    def visit_drop_index(self, drop: Any): ...
    def visit_drop_constraint(self, drop: Any): ...
    def define_constraint_match(self, constraint: Any): ...
    def visit_set_table_comment(self, create: Any): ...
    def visit_drop_table_comment(self, create: Any): ...
    def visit_set_column_comment(self, create: Any): ...

class MySQLTypeCompiler(compiler.GenericTypeCompiler):
    def visit_NUMERIC(self, type_: Any, **kw: Any): ...
    def visit_DECIMAL(self, type_: Any, **kw: Any): ...
    def visit_DOUBLE(self, type_: Any, **kw: Any): ...
    def visit_REAL(self, type_: Any, **kw: Any): ...
    def visit_FLOAT(self, type_: Any, **kw: Any): ...
    def visit_INTEGER(self, type_: Any, **kw: Any): ...
    def visit_BIGINT(self, type_: Any, **kw: Any): ...
    def visit_MEDIUMINT(self, type_: Any, **kw: Any): ...
    def visit_TINYINT(self, type_: Any, **kw: Any): ...
    def visit_SMALLINT(self, type_: Any, **kw: Any): ...
    def visit_BIT(self, type_: Any, **kw: Any): ...
    def visit_DATETIME(self, type_: Any, **kw: Any): ...
    def visit_DATE(self, type_: Any, **kw: Any): ...
    def visit_TIME(self, type_: Any, **kw: Any): ...
    def visit_TIMESTAMP(self, type_: Any, **kw: Any): ...
    def visit_YEAR(self, type_: Any, **kw: Any): ...
    def visit_TEXT(self, type_: Any, **kw: Any): ...
    def visit_TINYTEXT(self, type_: Any, **kw: Any): ...
    def visit_MEDIUMTEXT(self, type_: Any, **kw: Any): ...
    def visit_LONGTEXT(self, type_: Any, **kw: Any): ...
    def visit_VARCHAR(self, type_: Any, **kw: Any): ...
    def visit_CHAR(self, type_: Any, **kw: Any): ...
    def visit_NVARCHAR(self, type_: Any, **kw: Any): ...
    def visit_NCHAR(self, type_: Any, **kw: Any): ...
    def visit_VARBINARY(self, type_: Any, **kw: Any): ...
    def visit_JSON(self, type_: Any, **kw: Any): ...
    def visit_large_binary(self, type_: Any, **kw: Any): ...
    def visit_enum(self, type_: Any, **kw: Any): ...
    def visit_BLOB(self, type_: Any, **kw: Any): ...
    def visit_TINYBLOB(self, type_: Any, **kw: Any): ...
    def visit_MEDIUMBLOB(self, type_: Any, **kw: Any): ...
    def visit_LONGBLOB(self, type_: Any, **kw: Any): ...
    def visit_ENUM(self, type_: Any, **kw: Any): ...
    def visit_SET(self, type_: Any, **kw: Any): ...
    def visit_BOOLEAN(self, type_: Any, **kw: Any): ...

class MySQLIdentifierPreparer(compiler.IdentifierPreparer):
    reserved_words: Any = ...
    def __init__(
        self, dialect: Any, server_ansiquotes: bool = ..., **kw: Any
    ) -> None: ...

class MySQLDialect(default.DefaultDialect):
    name: str = ...
    supports_alter: bool = ...
    supports_native_boolean: bool = ...
    max_identifier_length: int = ...
    max_index_name_length: int = ...
    max_constraint_name_length: int = ...
    supports_native_enum: bool = ...
    supports_sequences: bool = ...
    sequences_optional: bool = ...
    supports_for_update_of: bool = ...
    supports_sane_rowcount: bool = ...
    supports_sane_multi_rowcount: bool = ...
    supports_multivalues_insert: bool = ...
    supports_comments: bool = ...
    inline_comments: bool = ...
    default_paramstyle: str = ...
    colspecs: Any = ...
    cte_follows_insert: bool = ...
    statement_compiler: Any = ...
    ddl_compiler: Any = ...
    type_compiler: Any = ...
    ischema_names: Any = ...
    preparer: Any = ...
    is_mariadb: bool = ...
    construct_arguments: Any = ...
    isolation_level: Any = ...
    def __init__(
        self,
        isolation_level: Optional[Any] = ...,
        json_serializer: Optional[Any] = ...,
        json_deserializer: Optional[Any] = ...,
        is_mariadb: Optional[Any] = ...,
        **kwargs: Any,
    ) -> None: ...
    def on_connect(self): ...
    def set_isolation_level(self, connection: Any, level: Any) -> None: ...
    def get_isolation_level(self, connection: Any): ...
    def do_begin_twophase(self, connection: Any, xid: Any) -> None: ...
    def do_prepare_twophase(self, connection: Any, xid: Any) -> None: ...
    def do_rollback_twophase(
        self,
        connection: Any,
        xid: Any,
        is_prepared: bool = ...,
        recover: bool = ...,
    ) -> None: ...
    def do_commit_twophase(
        self,
        connection: Any,
        xid: Any,
        is_prepared: bool = ...,
        recover: bool = ...,
    ) -> None: ...
    def do_recover_twophase(self, connection: Any): ...
    def is_disconnect(self, e: Any, connection: Any, cursor: Any): ...
    def has_table(
        self, connection: Any, table_name: Any, schema: Optional[Any] = ...
    ): ...
    def has_sequence(
        self, connection: Any, sequence_name: Any, schema: Optional[Any] = ...
    ): ...
    def get_sequence_names(
        self, connection: Any, schema: Optional[Any] = ..., **kw: Any
    ): ...
    identifier_preparer: Any = ...
    def initialize(self, connection: Any) -> None: ...
    def get_schema_names(self, connection: Any, **kw: Any): ...
    def get_table_names(
        self, connection: Any, schema: Optional[Any] = ..., **kw: Any
    ): ...
    def get_view_names(
        self, connection: Any, schema: Optional[Any] = ..., **kw: Any
    ): ...
    def get_table_options(
        self,
        connection: Any,
        table_name: Any,
        schema: Optional[Any] = ...,
        **kw: Any,
    ): ...
    def get_columns(
        self,
        connection: Any,
        table_name: Any,
        schema: Optional[Any] = ...,
        **kw: Any,
    ): ...
    def get_pk_constraint(
        self,
        connection: Any,
        table_name: Any,
        schema: Optional[Any] = ...,
        **kw: Any,
    ): ...
    def get_foreign_keys(
        self,
        connection: Any,
        table_name: Any,
        schema: Optional[Any] = ...,
        **kw: Any,
    ): ...
    def get_check_constraints(
        self,
        connection: Any,
        table_name: Any,
        schema: Optional[Any] = ...,
        **kw: Any,
    ): ...
    def get_table_comment(
        self,
        connection: Any,
        table_name: Any,
        schema: Optional[Any] = ...,
        **kw: Any,
    ): ...
    def get_indexes(
        self,
        connection: Any,
        table_name: Any,
        schema: Optional[Any] = ...,
        **kw: Any,
    ): ...
    def get_unique_constraints(
        self,
        connection: Any,
        table_name: Any,
        schema: Optional[Any] = ...,
        **kw: Any,
    ): ...
    def get_view_definition(
        self,
        connection: Any,
        view_name: Any,
        schema: Optional[Any] = ...,
        **kw: Any,
    ): ...

class _DecodingRow:
    rowproxy: Any = ...
    charset: Any = ...
    def __init__(self, rowproxy: Any, charset: Any) -> None: ...
    def __getitem__(self, index: Any): ...
    def __getattr__(self, attr: Any): ...



from typing import Any
from typing import Optional

from .base import _MSDate
from .base import _MSTime
from .base import MSDialect as MSDialect
from .base import VARBINARY as VARBINARY
from .pyodbc import _MSNumeric_pyodbc
from .pyodbc import MSExecutionContext_pyodbc as MSExecutionContext_pyodbc
from ...connectors.mxodbc import MxODBCConnector as MxODBCConnector

class _MSNumeric_mxodbc(_MSNumeric_pyodbc): ...

class _MSDate_mxodbc(_MSDate):
    def bind_processor(self, dialect: Any): ...

class _MSTime_mxodbc(_MSTime):
    def bind_processor(self, dialect: Any): ...

class _VARBINARY_mxodbc(VARBINARY):
    def bind_processor(self, dialect: Any): ...

class MSExecutionContext_mxodbc(MSExecutionContext_pyodbc): ...

class MSDialect_mxodbc(MxODBCConnector, MSDialect):
    execution_ctx_cls: Any = ...
    colspecs: Any = ...
    description_encoding: Any = ...
    def __init__(
        self, description_encoding: Optional[Any] = ..., **params: Any
    ) -> None: ...

dialect = MSDialect_mxodbc



from typing import Any

from .kinterbasdb import FBDialect_kinterbasdb as FBDialect_kinterbasdb
from ... import util as util

class FBDialect_fdb(FBDialect_kinterbasdb):
    def __init__(
        self, enable_rowcount: bool = ..., retaining: bool = ..., **kwargs: Any
    ) -> None: ...
    @classmethod
    def dbapi(cls): ...
    def create_connect_args(self, url: Any): ...

dialect = FBDialect_fdb



import os

from setuptools import setup


def find_stub_files():
    result = []
    for root, dirs, files in os.walk("sqlalchemy-stubs"):
        for file in files:
            if file.endswith(".pyi"):
                if os.path.sep in root:
                    sub_root = root.split(os.path.sep, 1)[-1]
                    file = os.path.join(sub_root, file)
                result.append(file)
    return result


setup(
    packages=["sqlalchemy-stubs"],
    package_data={"sqlalchemy-stubs": find_stub_files()},
)



from typing import Any

from .base import PGCompiler as PGCompiler
from .base import PGDialect as PGDialect
from .base import PGIdentifierPreparer as PGIdentifierPreparer
from .base import UUID as UUID
from .hstore import HSTORE as HSTORE
from .json import JSON as JSON
from .json import JSONB as JSONB
from ... import exc as exc
from ... import processors as processors
from ... import util as util
from ...sql.elements import Null as Null
from ...types import Numeric as Numeric

class _PGNumeric(Numeric):
    def bind_processor(self, dialect: Any) -> None: ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class _PGHStore(HSTORE):
    def bind_processor(self, dialect: Any): ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class _PGJSON(JSON):
    def bind_processor(self, dialect: Any): ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class _PGJSONB(JSONB):
    def bind_processor(self, dialect: Any): ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class _PGUUID(UUID):
    def bind_processor(self, dialect: Any): ...
    def result_processor(self, dialect: Any, coltype: Any): ...

class _PGCompiler(PGCompiler):
    def visit_mod_binary(self, binary: Any, operator: Any, **kw: Any): ...
    def post_process_text(self, text: Any): ...

class _PGIdentifierPreparer(PGIdentifierPreparer): ...

class PGDialect_pygresql(PGDialect):
    driver: str = ...
    statement_compiler: Any = ...
    preparer: Any = ...
    @classmethod
    def dbapi(cls): ...
    colspecs: Any = ...
    dbapi_version: Any = ...
    supports_unicode_statements: bool = ...
    supports_unicode_binds: bool = ...
    has_native_hstore: Any = ...
    has_native_json: Any = ...
    has_native_uuid: Any = ...
    def __init__(self, **kwargs: Any) -> None: ...
    def create_connect_args(self, url: Any): ...
    def is_disconnect(self, e: Any, connection: Any, cursor: Any): ...

dialect = PGDialect_pygresql



# fmt: off
from ...testing.provision import create_db as create_db
from ...testing.provision import drop_db as drop_db
from ...testing.provision import follower_url_from_main as follower_url_from_main
from ...testing.provision import log as log
from ...testing.provision import post_configure_engine as post_configure_engine
from ...testing.provision import run_reap_dbs as run_reap_dbs
from ...testing.provision import stop_test_class_outside_fixtures as stop_test_class_outside_fixtures
from ...testing.provision import temp_table_keyword_args as temp_table_keyword_args



from typing import Any
from typing import Deque
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import NamedTuple
from typing import NoReturn
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TypeVar
from typing import Union

from .elements import ClauseElement
from .visitors import ExtendedInternalTraversal
from .visitors import InternalTraversal
from .. import util
from ..util import HasMemoized
from ..util import langhelpers

_T = TypeVar("_T")
_KT = TypeVar("_KT")
_VT = TypeVar("_VT")
_CE = TypeVar("_CE", bound=ClauseElement)

SKIP_TRAVERSE: langhelpers._symbol
COMPARE_FAILED: bool
COMPARE_SUCCEEDED: bool
NO_CACHE: langhelpers._symbol
CACHE_IN_PLACE: langhelpers._symbol
CALL_GEN_CACHE_KEY: langhelpers._symbol
STATIC_CACHE_KEY: langhelpers._symbol
PROPAGATE_ATTRS: langhelpers._symbol
ANON_NAME: langhelpers._symbol

def compare(obj1: Any, obj2: Any, **kw: Any) -> bool: ...

class HasCacheKey: ...
class MemoizedHasCacheKey(HasCacheKey, HasMemoized): ...

class CacheKey(NamedTuple):
    key: Any
    bindparams: Any
    def to_offline_string(
        self, statement_cache: Any, statement: Any, parameters: Any
    ) -> str: ...
    def __eq__(self, other: object) -> bool: ...

class _CacheKey(ExtendedInternalTraversal):
    visit_has_cache_key: langhelpers._symbol = ...
    visit_clauseelement: langhelpers._symbol = ...
    visit_clauseelement_list: langhelpers._symbol = ...
    visit_annotations_key: langhelpers._symbol = ...
    visit_clauseelement_tuple: langhelpers._symbol = ...
    visit_string: langhelpers._symbol = ...
    visit_boolean: langhelpers._symbol = ...
    visit_operator: langhelpers._symbol = ...
    visit_plain_obj: langhelpers._symbol = ...
    visit_statement_hint_list: langhelpers._symbol = ...
    visit_type: langhelpers._symbol = ...
    visit_anon_name: langhelpers._symbol = ...
    visit_propagate_attrs: langhelpers._symbol = ...
    def visit_inspectable(
        self,
        attrname: str,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[str, Any]: ...
    def visit_string_list(
        self,
        attrname: Any,
        obj: Iterable[_T],
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[_T, ...]: ...
    def visit_multi(
        self,
        attrname: str,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[str, Any]: ...
    def visit_multi_list(
        self,
        attrname: str,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[str, Tuple[Any, ...]]: ...
    def visit_has_cache_key_tuples(
        self,
        attrname: str,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_has_cache_key_list(
        self,
        attrname: str,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_executable_options(
        self,
        attrname: str,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_inspectable_list(
        self,
        attrname: str,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_clauseelement_tuples(
        self,
        attrname: str,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_fromclause_ordered_set(
        self,
        attrname: str,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_clauseelement_unordered_set(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_named_ddl_element(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_prefix_sequence(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_setup_join_tuple(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_table_hint_list(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_plain_dict(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_dialect_options(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_string_clauseelement_dict(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_string_multi_dict(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_fromclause_canonical_column_collection(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_unknown_structure(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_dml_ordered_values(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_dml_values(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...
    def visit_dml_multi_values(
        self,
        attrname: Any,
        obj: Any,
        parent: Any,
        anon_map: Any,
        bindparams: Any,
    ) -> Tuple[Any, ...]: ...

class HasCopyInternals: ...

class _CopyInternals(InternalTraversal):
    def visit_clauseelement(
        self,
        attrname: Any,
        parent: Any,
        element: _CE,
        clone: Any = ...,
        **kw: Any,
    ) -> _CE: ...
    def visit_clauseelement_list(
        self,
        attrname: Any,
        parent: Any,
        element: Iterable[_CE],
        clone: Any = ...,
        **kw: Any,
    ) -> List[_CE]: ...
    def visit_clauseelement_tuple(
        self,
        attrname: Any,
        parent: Any,
        element: Iterable[_CE],
        clone: Any = ...,
        **kw: Any,
    ) -> Tuple[_CE, ...]: ...
    def visit_executable_options(
        self,
        attrname: Any,
        parent: Any,
        element: Iterable[_CE],
        clone: Any = ...,
        **kw: Any,
    ) -> Tuple[_CE, ...]: ...
    def visit_clauseelement_unordered_set(
        self,
        attrname: Any,
        parent: Any,
        element: Iterable[_CE],
        clone: Any = ...,
        **kw: Any,
    ) -> Set[_CE]: ...
    def visit_clauseelement_tuples(
        self,
        attrname: Any,
        parent: Any,
        element: Iterable[Iterable[_CE]],
        clone: Any = ...,
        **kw: Any,
    ) -> List[Tuple[_CE, ...]]: ...
    def visit_string_clauseelement_dict(
        self,
        attrname: Any,
        parent: Any,
        element: Mapping[_T, _CE],
        clone: Any = ...,
        **kw: Any,
    ) -> Dict[_T, _CE]: ...
    def visit_setup_join_tuple(
        self,
        attrname: Any,
        parent: Any,
        element: Iterable[Tuple[Any, Any, Any, Any]],
        clone: Any = ...,
        **kw: Any,
    ) -> Tuple[Tuple[Any, Any, Any, Any]]: ...
    def visit_dml_ordered_values(
        self,
        attrname: Any,
        parent: Any,
        element: Iterable[Tuple[Any, Any]],
        clone: Any = ...,
        **kw: Any,
    ) -> List[Tuple[Any, Any]]: ...
    def visit_dml_values(
        self,
        attrname: Any,
        parent: Any,
        element: Mapping[Any, Any],
        clone: Any = ...,
        **kw: Any,
    ) -> Dict[Any, Any]: ...
    def visit_dml_multi_values(
        self,
        attrname: Any,
        parent: Any,
        element: Iterable[Iterable[Any]],
        clone: Any = ...,
        **kw: Any,
    ) -> List[List[Union[List[Any], Dict[Any, Any], bool]]]: ...
    def visit_propagate_attrs(
        self,
        attrname: Any,
        parent: Any,
        element: _T,
        clone: Any = ...,
        **kw: Any,
    ) -> _T: ...

class _GetChildren(InternalTraversal):
    def visit_has_cache_key(
        self, element: Any, **kw: Any
    ) -> Tuple[Any, ...]: ...
    def visit_clauseelement(self, element: _CE, **kw: Any) -> Tuple[_CE]: ...
    def visit_clauseelement_list(self, element: _T, **kw: Any) -> _T: ...
    def visit_clauseelement_tuple(self, element: _T, **kw: Any) -> _T: ...
    def visit_clauseelement_tuples(
        self, element: Iterable[Iterable[_CE]], **kw: Any
    ) -> Iterator[_CE]: ...
    def visit_fromclause_canonical_column_collection(
        self, element: Any, **kw: Any
    ) -> Tuple[Any, ...]: ...
    def visit_string_clauseelement_dict(
        self, element: Mapping[Any, _CE], **kw: Any
    ) -> Iterable[_CE]: ...
    def visit_fromclause_ordered_set(self, element: _T, **kw: Any) -> _T: ...
    def visit_clauseelement_unordered_set(
        self, element: _CE, **kw: Any
    ) -> _CE: ...
    def visit_setup_join_tuple(
        self, element: Iterable[Tuple[Any, Any, Any, Any]], **kw: Any
    ) -> Iterator[Any]: ...
    def visit_dml_ordered_values(
        self, element: Iterable[Tuple[Any, Any]], **kw: Any
    ) -> Iterator[Any]: ...
    def visit_dml_values(
        self, element: Mapping[Any, Any], **kw: Any
    ) -> Iterator[Any]: ...
    def visit_dml_multi_values(
        self, element: Any, **kw: Any
    ) -> Tuple[Any, ...]: ...
    def visit_propagate_attrs(
        self, element: Any, **kw: Any
    ) -> Tuple[Any, ...]: ...

class anon_map(Dict[_KT, _VT]):
    index: int = ...
    def __init__(self) -> None: ...
    def __missing__(self, key: Any) -> str: ...

class TraversalComparatorStrategy(InternalTraversal, util.MemoizedSlots):
    stack: Deque[Tuple[Any, Any]] = ...
    cache: Set[Tuple[Any, Any]] = ...
    def __init__(self) -> None: ...
    def compare(self, obj1: Any, obj2: Any, **kw: Any) -> bool: ...
    def compare_inner(self, obj1: Any, obj2: Any, **kw: Any) -> bool: ...
    def visit_has_cache_key(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> Optional[int]: ...
    def visit_propagate_attrs(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_has_cache_key_list(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> Optional[int]: ...
    def visit_executable_options(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> Optional[int]: ...
    def visit_clauseelement(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> None: ...
    def visit_fromclause_canonical_column_collection(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> None: ...
    def visit_fromclause_derived_column_collection(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> None: ...
    def visit_string_clauseelement_dict(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> Optional[int]: ...
    def visit_clauseelement_tuples(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> Optional[int]: ...
    def visit_clauseelement_list(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> None: ...
    def visit_clauseelement_tuple(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> None: ...
    def visit_clauseelement_unordered_set(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_fromclause_ordered_set(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> None: ...
    def visit_string(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_string_list(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_anon_name(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_boolean(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_operator(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_type(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_plain_dict(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_dialect_options(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_annotations_key(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_plain_obj(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_named_ddl_element(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> Union[int, bool]: ...
    def visit_prefix_sequence(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> Optional[int]: ...
    def visit_setup_join_tuple(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> Optional[int]: ...
    def visit_table_hint_list(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> Optional[int]: ...
    def visit_statement_hint_list(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> bool: ...
    def visit_unknown_structure(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> NoReturn: ...
    def visit_dml_ordered_values(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> Optional[int]: ...
    def visit_dml_values(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> Optional[int]: ...
    def visit_dml_multi_values(
        self,
        attrname: Any,
        left_parent: Any,
        left: Any,
        right_parent: Any,
        right: Any,
        **kw: Any,
    ) -> Optional[int]: ...
    def compare_clauselist(
        self, left: Any, right: Any, **kw: Any
    ) -> Union[List[str], int]: ...
    def compare_binary(
        self, left: Any, right: Any, **kw: Any
    ) -> Union[List[str], int]: ...
    def compare_bindparam(
        self, left: Any, right: Any, **kw: Any
    ) -> List[str]: ...

class ColIdentityComparatorStrategy(TraversalComparatorStrategy):
    def compare_column_element(
        self,
        left: Any,
        right: Any,
        use_proxies: bool = ...,
        equivalents: Any = ...,
        **kw: Any,
    ) -> int: ...
    def compare_column(self, left: Any, right: Any, **kw: Any) -> int: ...
    def compare_label(self, left: Any, right: Any, **kw: Any) -> int: ...
    def compare_table(self, left: Any, right: Any, **kw: Any) -> int: ...



from typing import Any

from .types import _StringType
from ... import exc as exc
from ... import sql as sql
from ... import util as util
from ...sql import sqltypes as sqltypes
from ...sql.base import NO_ARG as NO_ARG

class ENUM(sqltypes.NativeForEmulated, sqltypes.Enum, _StringType):
    __visit_name__: str = ...
    native_enum: bool = ...
    def __init__(self, *enums: Any, **kw: Any) -> None: ...
    @classmethod
    def adapt_emulated_to_native(cls, impl: Any, **kw: Any): ...

class SET(_StringType):
    __visit_name__: str = ...
    retrieve_as_bitwise: Any = ...
    values: Any = ...
    def __init__(self, *values: Any, **kw: Any) -> None: ...
    def column_expression(self, colexpr: Any): ...
    def result_processor(self, dialect: Any, coltype: Any): ...
    def bind_processor(self, dialect: Any): ...
    def adapt(self, impltype: Any, **kw: Any): ...



from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import Union

from typing_extensions import Protocol
from typing_extensions import TypedDict

from .base import Connection
from .base import Engine
from .cursor import CursorFetchStrategy
from .cursor import CursorResult
from .url import URL
from ..exc import StatementError
from ..sql.compiler import Compiled as Compiled
from ..sql.compiler import IdentifierPreparer
from ..sql.compiler import TypeCompiler as TypeCompiler
from ..sql.schema import Column
from ..sql.type_api import TypeEngine

class _DBAPIConnectionBase(Protocol):
    def close(self) -> None: ...
    def commit(self) -> None: ...
    def cursor(self) -> _DBAPICursor: ...

class _DBAPIConnectionTransactions(_DBAPIConnectionBase, Protocol):
    def rollback(self) -> None: ...

_DBAPIConnection = Union[_DBAPIConnectionBase, _DBAPIConnectionTransactions]

class _DBAPICursorBase(Protocol):
    @property
    def description(self) -> Any: ...
    @property
    def rowcount(self) -> int: ...
    arraysize: int = ...
    def close(self) -> None: ...
    def execute(self, operation: Any, parameters: Any) -> Any: ...
    def executemany(
        self, operation: Any, parameters: Sequence[Any]
    ) -> Any: ...
    def fetchone(self) -> Optional[Any]: ...
    def fetchmany(self, size: int = ...) -> Sequence[Any]: ...
    def fetchall(self) -> Sequence[Any]: ...
    def setinputsizes(self, sizes: Sequence[Any]) -> None: ...
    def setoutputsize(self, size: Any, column: Any) -> None: ...

class _DBAPICursorStoredProcedures(_DBAPICursorBase, Protocol):
    def callproc(
        self, procname: str, parameters: Sequence[Any] = ...
    ) -> Any: ...

class _DBAPICursorMultiResultSets(_DBAPICursorBase, Protocol):
    def nextset(self) -> Optional[bool]: ...

_DBAPICursor = Union[
    _DBAPICursorBase, _DBAPICursorStoredProcedures, _DBAPICursorMultiResultSets
]

class _OnConnect(Protocol):
    def __call__(self, connection: _DBAPIConnection) -> None: ...

class _PKConstraintDictBase(TypedDict):
    constrained_columns: List[str]

class _PKConstraintDict(_PKConstraintDictBase, total=False):
    name: str

class _ForeignKeyDict(TypedDict):
    name: str
    constrained_columns: List[str]
    referred_schema: str
    referred_table: str
    referred_columns: List[str]

class _IndexDict(TypedDict):
    name: str
    column_names: List[str]
    unique: bool

class _TableCommentDict(TypedDict):
    text: str

class Dialect:
    name: str = ...
    driver: str = ...
    positional: bool = ...
    paramstyle: str = ...
    encoding: str = ...
    statement_compiler: Type[Compiled] = ...
    ddl_compiler: Type[Compiled] = ...
    server_version_info: Optional[Tuple[Any, ...]] = ...
    default_schema_name: Optional[str] = ...
    execution_ctx_cls: Type[ExecutionContext] = ...
    execute_sequence_format: Union[
        Type[Tuple[Any, ...]], Type[List[Any]]
    ] = ...
    preparer: Type[IdentifierPreparer] = ...
    supports_alter: bool = ...
    max_identifier_length: int = ...
    supports_sane_rowcount: bool = ...
    supports_sane_multi_rowcount: bool = ...
    preexecute_autoincrement_sequences: bool = ...
    implicit_returning: bool = ...
    colspecs: Dict[Type[TypeEngine[Any]], Type[TypeEngine[Any]]] = ...
    supports_sequences: bool = ...
    sequences_optional: bool = ...
    supports_native_enum: bool = ...
    supports_native_boolean: bool = ...
    dbapi_exception_translation_map: Dict[str, str] = ...
    supports_statement_cache: bool = ...
    def create_connect_args(
        self, url: URL
    ) -> Tuple[Sequence[Any], Mapping[str, Any]]: ...
    @classmethod
    def type_descriptor(cls, typeobj: Type[TypeEngine[Any]]) -> Any: ...
    def initialize(self, connection: Connection) -> None: ...
    def get_columns(
        self,
        connection: Connection,
        table_name: str,
        schema: Optional[str] = ...,
        **kw: Any,
    ) -> List[Dict[str, Any]]: ...
    def get_pk_constraint(
        self,
        connection: Connection,
        table_name: str,
        schema: Optional[str] = ...,
        **kw: Any,
    ) -> _PKConstraintDict: ...
    def get_foreign_keys(
        self,
        connection: Connection,
        table_name: str,
        schema: Optional[str] = ...,
        **kw: Any,
    ) -> List[_ForeignKeyDict]: ...
    def get_table_names(
        self, connection: Connection, schema: Optional[str] = ..., **kw: Any
    ) -> List[str]: ...
    def get_temp_table_names(
        self, connection: Connection, schema: Optional[str] = ..., **kw: Any
    ) -> List[str]: ...
    def get_view_names(
        self, connection: Connection, schema: Optional[str] = ..., **kw: Any
    ) -> List[str]: ...
    def get_sequence_names(
        self, connection: Connection, schema: Optional[str] = ..., **kw: Any
    ) -> List[str]: ...
    def get_temp_view_names(
        self, connection: Connection, schema: Optional[str] = ..., **kw: Any
    ) -> List[str]: ...
    def get_view_definition(
        self,
        connection: Connection,
        view_name: str,
        schema: Optional[str] = ...,
        **kw: Any,
    ) -> Any: ...
    def get_indexes(
        self,
        connection: Connection,
        table_name: str,
        schema: Optional[str] = ...,
        **kw: Any,
    ) -> List[_IndexDict]: ...
    def get_unique_constraints(
        self,
        connection: Connection,
        table_name: str,
        schema: Optional[str] = ...,
        **kw: Any,
    ) -> List[Dict[str, Any]]: ...
    def get_check_constraints(
        self,
        connection: Connection,
        table_name: str,
        schema: Optional[str] = ...,
        **kw: Any,
    ) -> List[Dict[str, Any]]: ...
    def get_table_comment(
        self,
        connection: Connection,
        table_name: str,
        schema: Optional[str] = ...,
        **kw: Any,
    ) -> _TableCommentDict: ...
    def normalize_name(self, name: Optional[str]) -> Optional[str]: ...
    def denormalize_name(self, name: Optional[str]) -> Optional[str]: ...
    def has_table(
        self,
        connection: Connection,
        table_name: str,
        schema: Optional[str] = ...,
        **kw: Any,
    ) -> bool: ...
    def has_index(
        self,
        connection: Connection,
        table_name: str,
        index_name: str,
        schema: Optional[str] = ...,
    ) -> bool: ...
    def has_sequence(
        self,
        connection: Connection,
        sequence_name: str,
        schema: Optional[str] = ...,
        **kw: Any,
    ) -> bool: ...
    def do_begin(self, dbapi_connection: _DBAPIConnection) -> None: ...
    def do_rollback(self, dbapi_connection: _DBAPIConnection) -> None: ...
    def do_commit(self, dbapi_connection: _DBAPIConnection) -> None: ...
    def do_close(self, dbapi_connection: _DBAPIConnection) -> None: ...
    def do_set_input_sizes(
        self,
        cursor: _DBAPICursor,
        list_of_tuples: Tuple[str, Any, TypeEngine[Any]],
        context: ExecutionContext,
    ) -> None: ...
    def create_xid(self) -> str: ...
    def do_savepoint(self, connection: Connection, name: str) -> None: ...
    def do_rollback_to_savepoint(
        self, connection: Connection, name: str
    ) -> None: ...
    def do_release_savepoint(
        self, connection: Connection, name: str
    ) -> None: ...
    def do_begin_twophase(self, connection: Connection, xid: str) -> None: ...
    def do_prepare_twophase(
        self, connection: Connection, xid: str
    ) -> None: ...
    def do_rollback_twophase(
        self,
        connection: Connection,
        xid: str,
        is_prepared: bool = ...,
        recover: bool = ...,
    ) -> None: ...
    def do_commit_twophase(
        self,
        connection: Connection,
        xid: str,
        is_prepared: bool = ...,
        recover: bool = ...,
    ) -> None: ...
    def do_recover_twophase(self, connection: Connection) -> None: ...
    def do_executemany(
        self,
        cursor: _DBAPICursor,
        statement: Any,
        parameters: Any,
        context: Optional[ExecutionContext] = ...,
    ) -> None: ...
    def do_execute(
        self,
        cursor: _DBAPICursor,
        statement: Any,
        parameters: Any,
        context: Optional[ExecutionContext] = ...,
    ) -> None: ...
    def do_execute_no_params(
        self,
        cursor: _DBAPICursor,
        statement: Any,
        parameters: Any,
        context: Optional[ExecutionContext] = ...,
    ) -> None: ...
    def is_disconnect(
        self, e: Any, connection: Connection, cursor: _DBAPICursor
    ) -> bool: ...
    def connect(self, *cargs: Any, **cparams: Any) -> _DBAPIConnection: ...
    def on_connect_url(self, url: URL) -> Optional[_OnConnect]: ...
    def on_connect(self) -> Optional[_OnConnect]: ...
    def reset_isolation_level(self, dbapi_conn: _DBAPIConnection) -> None: ...
    def set_isolation_level(
        self, dbapi_conn: _DBAPIConnection, level: Any
    ) -> None: ...
    def get_isolation_level(self, dbapi_conn: _DBAPIConnection) -> Any: ...
    def get_default_isolation_level(
        self, dbapi_conn: _DBAPIConnection
    ) -> Any: ...
    @classmethod
    def get_dialect_cls(cls, url: URL) -> Dialect: ...
    @classmethod
    def load_provisioning(cls) -> None: ...
    @classmethod
    def engine_created(cls, engine: Engine) -> None: ...

class CreateEnginePlugin:
    url: URL = ...
    def __init__(self, url: URL, kwargs: Dict[str, Any]) -> None: ...
    def update_url(self, url: URL) -> URL: ...
    def handle_dialect_kwargs(
        self, dialect_cls: Type[Dialect], dialect_args: Dict[str, Any]
    ) -> None: ...
    def handle_pool_kwargs(
        self, pool_cls: Any, pool_args: Dict[str, Any]
    ) -> None: ...
    def engine_created(self, engine: Engine) -> None: ...

class ExecutionContext:
    connection: Connection = ...
    root_connection: Connection = ...
    dialect: Dialect = ...
    cursor: _DBAPICursor = ...
    compiled: Optional[Compiled] = ...
    statement: Optional[str] = ...
    parameters: Any = ...
    isinsert: bool = ...
    isupdate: bool = ...
    should_autocommit: bool = ...
    prefetch_cols: Sequence[Column[Any]] = ...
    postfetch_cols: Sequence[Column[Any]] = ...
    def create_cursor(self) -> _DBAPICursor: ...
    def pre_exec(self) -> None: ...
    def get_out_parameter_values(
        self, out_param_names: Sequence[str]
    ) -> Sequence[Any]: ...
    def post_exec(self) -> None: ...
    def get_result_cursor_strategy(
        self, result: Any
    ) -> CursorFetchStrategy: ...
    def handle_dbapi_exception(self, e: Any) -> None: ...
    def should_autocommit_text(self, statement: str) -> bool: ...
    def lastrow_has_defaults(self) -> bool: ...
    def get_rowcount(self) -> int: ...

class Connectable:
    def connect(self, **kwargs: Any) -> Connection: ...
    engine: Optional[Engine] = ...
    def execute(
        self, object_: Any, *multiparams: Any, **params: Any
    ) -> CursorResult: ...
    def scalar(
        self, object_: Any, *multiparams: Any, **params: Any
    ) -> Any: ...

class ExceptionContext:
    connection: Optional[Connection] = ...
    engine: Engine = ...
    cursor: Optional[_DBAPICursor] = ...
    statement: Optional[str] = ...
    parameters: Optional[Any] = ...
    original_exception: BaseException = ...
    sqlalchemy_exception: Optional[StatementError] = ...
    chained_exception: Optional[BaseException] = ...
    execution_context: Optional[ExecutionContext] = ...
    is_disconnect: bool = ...
    invalidate_pool_on_disconnect: bool = ...



from typing import Any

from .. import event as event
from .. import types as types
from ..orm import Mapper as Mapper
from ..orm import mapper as mapper
from ..orm import object_mapper as object_mapper
from ..orm.attributes import flag_modified as flag_modified
from ..sql.base import SchemaEventTarget as SchemaEventTarget
from ..util import memoized_property as memoized_property

class MutableBase:
    @classmethod
    def coerce(cls, key: Any, value: Any) -> None: ...

class Mutable(MutableBase):
    def changed(self) -> None: ...
    @classmethod
    def associate_with_attribute(cls, attribute: Any) -> None: ...
    @classmethod
    def associate_with(cls, sqltype: Any) -> None: ...
    @classmethod
    def as_mutable(cls, sqltype: Any): ...

class MutableComposite(MutableBase):
    def changed(self) -> None: ...

class MutableDict(Mutable, dict):
    def __setitem__(self, key: Any, value: Any) -> None: ...
    def setdefault(self, key: Any, value: Any): ...
    def __delitem__(self, key: Any) -> None: ...
    def update(self, *a: Any, **kw: Any) -> None: ...
    def pop(self, *arg: Any): ...
    def popitem(self): ...
    def clear(self) -> None: ...
    @classmethod
    def coerce(cls, key: Any, value: Any): ...

class MutableList(Mutable, list):
    def __reduce_ex__(self, proto: Any): ...
    def __setitem__(self, index: Any, value: Any) -> None: ...
    def __setslice__(self, start: Any, end: Any, value: Any) -> None: ...
    def __delitem__(self, index: Any) -> None: ...
    def __delslice__(self, start: Any, end: Any) -> None: ...
    def pop(self, *arg: Any): ...
    def append(self, x: Any) -> None: ...
    def extend(self, x: Any) -> None: ...
    def __iadd__(self, x: Any): ...
    def insert(self, i: Any, x: Any) -> None: ...
    def remove(self, i: Any) -> None: ...
    def clear(self) -> None: ...
    def sort(self, **kw: Any) -> None: ...
    def reverse(self) -> None: ...
    @classmethod
    def coerce(cls, index: Any, value: Any): ...

class MutableSet(Mutable, set):
    def update(self, *arg: Any) -> None: ...
    def intersection_update(self, *arg: Any) -> None: ...
    def difference_update(self, *arg: Any) -> None: ...
    def symmetric_difference_update(self, *arg: Any) -> None: ...
    def __ior__(self, other: Any): ...
    def __iand__(self, other: Any): ...
    def __ixor__(self, other: Any): ...
    def __isub__(self, other: Any): ...
    def add(self, elem: Any) -> None: ...
    def remove(self, elem: Any) -> None: ...
    def discard(self, elem: Any) -> None: ...
    def pop(self, *arg: Any): ...
    def clear(self) -> None: ...
    @classmethod
    def coerce(cls, index: Any, value: Any): ...
    def __reduce_ex__(self, proto: Any): ...



from typing import Any

from ... import types as sqltypes

class RangeOperators:
    class comparator_factory(sqltypes.Concatenable.Comparator):
        def __ne__(self, other: Any) -> Any: ...
        def contains(self, other: Any, **kw: Any): ...
        def contained_by(self, other: Any): ...
        def overlaps(self, other: Any): ...
        def strictly_left_of(self, other: Any): ...
        __lshift__: Any = ...
        def strictly_right_of(self, other: Any): ...
        __rshift__: Any = ...
        def not_extend_right_of(self, other: Any): ...
        def not_extend_left_of(self, other: Any): ...
        def adjacent_to(self, other: Any): ...
        def __add__(self, other: Any): ...

class INT4RANGE(RangeOperators, sqltypes.TypeEngine):
    __visit_name__: str = ...

class INT8RANGE(RangeOperators, sqltypes.TypeEngine):
    __visit_name__: str = ...

class NUMRANGE(RangeOperators, sqltypes.TypeEngine):
    __visit_name__: str = ...

class DATERANGE(RangeOperators, sqltypes.TypeEngine):
    __visit_name__: str = ...

class TSRANGE(RangeOperators, sqltypes.TypeEngine):
    __visit_name__: str = ...

class TSTZRANGE(RangeOperators, sqltypes.TypeEngine):
    __visit_name__: str = ...



from typing import Any
from typing import Generic
from typing import Optional
from typing import TypeVar

from .base import _MappedAttribute as _MappedAttribute
from .base import EXT_CONTINUE as EXT_CONTINUE
from .base import EXT_SKIP as EXT_SKIP
from .base import EXT_STOP as EXT_STOP
from .base import InspectionAttr as InspectionAttr
from .base import InspectionAttrInfo as InspectionAttrInfo  # noqa
from .base import MANYTOMANY as MANYTOMANY
from .base import MANYTOONE as MANYTOONE
from .base import NOT_EXTENSION as NOT_EXTENSION
from .base import ONETOMANY as ONETOMANY
from .mapper import Mapper
from .util import AliasedInsp
from .. import util
from ..sql import operators
from ..sql import roles
from ..sql.base import ExecutableOption
from ..sql.traversals import HasCacheKey

class ORMStatementRole(roles.CoerceTextStatementRole): ...
class ORMColumnsClauseRole(roles.ColumnsClauseRole): ...
class ORMEntityColumnsClauseRole(ORMColumnsClauseRole): ...
class ORMFromClauseRole(roles.StrictFromClauseRole): ...

_T = TypeVar("_T")

class MapperProperty(
    HasCacheKey,
    _MappedAttribute,
    InspectionAttr,
    util.MemoizedSlots,
    Generic[_T],
):
    cascade: Any = ...
    is_property: bool = ...
    def setup(
        self,
        context: Any,
        query_entity: Any,
        path: Any,
        adapter: Any,
        **kwargs: Any,
    ) -> None: ...
    def create_row_processor(
        self,
        context: Any,
        query_entity: Any,
        path: Any,
        mapper: Any,
        result: Any,
        adapter: Any,
        populators: Any,
    ) -> None: ...
    def cascade_iterator(
        self,
        type_: Any,
        state: Any,
        dict_: Any,
        visited_states: Any,
        halt_on: Optional[Any] = ...,
    ) -> None: ...
    parent: Any = ...
    def set_parent(self, parent: Any, init: Any) -> None: ...
    def instrument_class(self, mapper: Any) -> None: ...
    def __init__(self) -> None: ...
    def init(self) -> None: ...
    @property
    def class_attribute(self): ...
    def do_init(self) -> None: ...
    def post_instrument_class(self, mapper: Any) -> None: ...
    def merge(
        self,
        session: Any,
        source_state: Any,
        source_dict: Any,
        dest_state: Any,
        dest_dict: Any,
        load: Any,
        _recursive: Any,
        _resolve_conflict_map: Any,
    ) -> None: ...

class PropComparator(operators.ColumnOperators):
    prop: Any = ...
    def __init__(
        self,
        prop: MapperProperty,
        parentmapper: Mapper,
        adapt_to_entity: Optional[AliasedInsp] = ...,
    ) -> None: ...
    def __clause_element__(self) -> None: ...
    def adapt_to_entity(self, adapt_to_entity: Any): ...
    @property
    def adapter(self): ...
    @property
    def info(self): ...
    @staticmethod
    def any_op(a: Any, b: Any, **kwargs: Any): ...
    @staticmethod
    def has_op(a: Any, b: Any, **kwargs: Any): ...
    @staticmethod
    def of_type_op(a: Any, class_: Any): ...
    def of_type(self, class_: Any): ...
    def and_(self, *criteria: Any): ...
    def any(self, criterion: Optional[Any] = ..., **kwargs: Any): ...
    def has(self, criterion: Optional[Any] = ..., **kwargs: Any): ...

class StrategizedProperty(MapperProperty[_T]):
    inherit_cache: bool = ...
    strategy_wildcard_key: Any = ...
    def setup(
        self,
        context: Any,
        query_entity: Any,
        path: Any,
        adapter: Any,
        **kwargs: Any,
    ) -> None: ...
    def create_row_processor(
        self,
        context: Any,
        query_entity: Any,
        path: Any,
        mapper: Any,
        result: Any,
        adapter: Any,
        populators: Any,
    ) -> None: ...
    strategy: Any = ...
    def do_init(self) -> None: ...
    def post_instrument_class(self, mapper: Any) -> None: ...
    @classmethod
    def strategy_for(cls, **kw: Any): ...

class ORMOption(ExecutableOption):
    propagate_to_loaders: bool = ...

class LoaderOption(ORMOption):
    def process_compile_state(self, compile_state: Any) -> None: ...

class CriteriaOption(ORMOption):
    def process_compile_state(self, compile_state: Any) -> None: ...
    def get_global_criteria(self, attributes: Any) -> None: ...

class UserDefinedOption(ORMOption):
    propagate_to_loaders: bool = ...
    payload: Any = ...
    def __init__(self, payload: Optional[Any] = ...) -> None: ...

class MapperOption(ORMOption):
    propagate_to_loaders: bool = ...
    def process_query(self, query: Any) -> None: ...
    def process_query_conditionally(self, query: Any) -> None: ...

class LoaderStrategy:
    parent_property: Any = ...
    is_class_level: bool = ...
    parent: Any = ...
    key: Any = ...
    strategy_key: Any = ...
    strategy_opts: Any = ...
    def __init__(self, parent: Any, strategy_key: Any) -> None: ...
    def init_class_attribute(self, mapper: Any) -> None: ...
    def setup_query(
        self,
        compile_state: Any,
        query_entity: Any,
        path: Any,
        loadopt: Any,
        adapter: Any,
        **kwargs: Any,
    ) -> None: ...
    def create_row_processor(
        self,
        context: Any,
        query_entity: Any,
        path: Any,
        loadopt: Any,
        mapper: Any,
        result: Any,
        adapter: Any,
        populators: Any,
    ) -> None: ...



from typing import Any
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

from typing_extensions import Protocol

from . import interfaces
from .session import _SessionProtocol
from .state import InstanceState
from ..sql.annotation import SupportsCloneAnnotations
from ..sql.base import Executable
from ..sql.elements import BooleanClauseList
from ..sql.elements import GroupedElement
from ..sql.elements import Label
from ..sql.selectable import _SelectFromElements
from ..sql.selectable import Alias
from ..sql.selectable import CTE
from ..sql.selectable import Exists
from ..sql.selectable import HasHints
from ..sql.selectable import HasPrefixes
from ..sql.selectable import HasSuffixes
from ..sql.selectable import ScalarSelect
from ..sql.selectable import Select
from ..sql.selectable import SelectBase
from ..sql.selectable import SelectStatementGrouping
from ..util.langhelpers import _symbol

_T = TypeVar("_T")
_TQuery = TypeVar("_TQuery", bound=Query[Any])

class _WithTransformationFn(Protocol[_TQuery]):
    def __call__(self, __query: _TQuery) -> _TQuery: ...

class Query(
    _SelectFromElements,
    SupportsCloneAnnotations,
    HasPrefixes,
    HasSuffixes,
    HasHints,
    Executable,
    Generic[_T],
):
    load_options: Any = ...
    session: Optional[_SessionProtocol] = ...
    def __init__(
        self,
        entities: Iterable[Any],
        session: Optional[_SessionProtocol] = ...,
    ) -> None: ...
    @property
    def statement(self) -> Union[Select, FromStatement]: ...
    def subquery(
        self,
        name: Optional[str] = ...,
        with_labels: bool = ...,
        reduce_columns: bool = ...,
    ) -> Alias: ...
    def cte(self, name: Optional[str] = ..., recursive: bool = ...) -> CTE: ...
    def label(self, name: str) -> Label[Any]: ...
    def as_scalar(self) -> ScalarSelect: ...
    def scalar_subquery(self) -> ScalarSelect: ...
    @property
    def selectable(self) -> Union[Select, FromStatement]: ...
    def __clause_element__(self) -> Union[Select, FromStatement]: ...
    def only_return_tuples(self: _TQuery, value: bool) -> _TQuery: ...
    @property
    def is_single_entity(self) -> bool: ...
    def enable_eagerloads(self: _TQuery, value: bool) -> _TQuery: ...
    def with_labels(self: _TQuery) -> _TQuery: ...
    def apply_labels(self: _TQuery) -> _TQuery: ...
    @property
    def get_label_style(self) -> _symbol: ...
    def set_label_style(self: _TQuery, style: _symbol) -> _TQuery: ...
    def enable_assertions(self: _TQuery, value: bool) -> _TQuery: ...
    @property
    def whereclause(self) -> BooleanClauseList[Any]: ...
    def with_polymorphic(
        self: _TQuery,
        cls_or_mappers: Any,
        selectable: Optional[Any] = ...,
        polymorphic_on: Optional[Any] = ...,
    ) -> _TQuery: ...
    def yield_per(self: _TQuery, count: int) -> _TQuery: ...
    def get(self, ident: Any) -> Optional[_T]: ...
    @property
    def lazy_loaded_from(self) -> InstanceState: ...
    def correlate(self: _TQuery, *fromclauses: Any) -> _TQuery: ...
    def autoflush(self: _TQuery, setting: Any) -> _TQuery: ...
    def populate_existing(self: _TQuery) -> _TQuery: ...
    def with_parent(
        self: _TQuery,
        instance: Any,
        property: Optional[Any] = ...,
        from_entity: Optional[Any] = ...,
    ) -> _TQuery: ...
    def add_entity(
        self: _TQuery, entity: Any, alias: Optional[Any] = ...
    ) -> _TQuery: ...
    def with_session(
        self: _TQuery, session: Optional[_SessionProtocol]
    ) -> _TQuery: ...
    def from_self(self: _TQuery, *entities: Any) -> _TQuery: ...
    def values(self, *columns: Any) -> Iterator[Tuple[Any, ...]]: ...
    def value(self, column: Any) -> Optional[Any]: ...
    def with_entities(self: _TQuery, *entities: Any) -> _TQuery: ...
    def add_columns(self: _TQuery, *column: Any) -> _TQuery: ...
    def add_column(self: _TQuery, column: Any) -> _TQuery: ...
    def options(self: _TQuery, *args: Any) -> _TQuery: ...
    def with_transformation(
        self: _TQuery, fn: _WithTransformationFn[_TQuery]
    ) -> _TQuery: ...
    def get_execution_options(self) -> Any: ...
    def execution_options(self: _TQuery, **kwargs: Any) -> _TQuery: ...
    def with_for_update(
        self: _TQuery,
        read: bool = ...,
        nowait: bool = ...,
        of: Optional[Any] = ...,
        skip_locked: bool = ...,
        key_share: bool = ...,
    ) -> _TQuery: ...
    def params(self: _TQuery, *args: Any, **kwargs: Any) -> _TQuery: ...
    def where(self: _TQuery, *criterion: Any) -> _TQuery: ...
    def filter(self: _TQuery, *criterion: Any) -> _TQuery: ...
    def filter_by(self: _TQuery, **kwargs: Any) -> _TQuery: ...
    def order_by(self: _TQuery, *clauses: Any) -> _TQuery: ...
    def group_by(self: _TQuery, *clauses: Any) -> _TQuery: ...
    def having(self: _TQuery, criterion: Any) -> _TQuery: ...
    def union(self: _TQuery, *q: Any) -> _TQuery: ...
    def union_all(self: _TQuery, *q: Any) -> _TQuery: ...
    def intersect(self: _TQuery, *q: Any) -> _TQuery: ...
    def intersect_all(self: _TQuery, *q: Any) -> _TQuery: ...
    def except_(self: _TQuery, *q: Any) -> _TQuery: ...
    def except_all(self: _TQuery, *q: Any) -> _TQuery: ...
    def join(
        self: _TQuery, target: Any, *props: Any, **kwargs: Any
    ) -> _TQuery: ...
    def outerjoin(
        self: _TQuery, target: Any, *props: Any, **kwargs: Any
    ) -> _TQuery: ...
    def reset_joinpoint(self: _TQuery) -> _TQuery: ...
    def select_from(self: _TQuery, *from_obj: Any) -> _TQuery: ...
    def select_entity_from(self: _TQuery, from_obj: Any) -> _TQuery: ...
    def __getitem__(self, item: Any) -> Any: ...
    def slice(self: _TQuery, start: Any, stop: Any) -> _TQuery: ...
    def limit(self: _TQuery, limit: Any) -> _TQuery: ...
    def offset(self: _TQuery, offset: Any) -> _TQuery: ...
    def distinct(self: _TQuery, *expr: Any) -> _TQuery: ...
    def all(self) -> List[_T]: ...
    def from_statement(self: _TQuery, statement: Any) -> _TQuery: ...
    def first(self) -> Optional[_T]: ...
    def one_or_none(self) -> Optional[_T]: ...
    def one(self) -> _T: ...
    def scalar(self) -> Any: ...  # type: ignore[override]
    def __iter__(self) -> Iterator[_T]: ...
    @property
    def column_descriptions(self) -> List[Any]: ...
    def instances(
        self, result_proxy: Any, context: Optional[Any] = ...
    ) -> Any: ...
    def merge_result(self, iterator: Any, load: bool = ...) -> Any: ...
    def exists(self) -> Exists: ...
    def count(self) -> int: ...
    def delete(self, synchronize_session: Union[str, bool] = ...) -> int: ...
    def update(
        self,
        values: Any,
        synchronize_session: Union[str, bool] = ...,
        update_args: Optional[Any] = ...,
    ) -> int: ...

class FromStatement(GroupedElement, SelectBase, Executable):
    __visit_name__: str = ...
    element: Any = ...
    def __init__(self, entities: Iterable[Any], element: Any) -> None: ...
    def get_label_style(self) -> _symbol: ...
    def set_label_style(
        self, label_style: _symbol
    ) -> SelectStatementGrouping: ...
    def get_children(self, **kw: Any) -> Iterable[Any]: ...  # type: ignore[override]

class AliasOption(interfaces.LoaderOption):
    def __init__(self, alias: Any) -> None: ...
    def process_compile_state(self, compile_state: Any) -> None: ...

class BulkUD:
    query: Query[Any] = ...
    mapper: Any = ...
    def __init__(self, query: Query[Any]) -> None: ...
    @property
    def session(self) -> Optional[_SessionProtocol]: ...

class BulkUpdate(BulkUD):
    values: Any = ...
    update_kwargs: Any = ...
    def __init__(
        self, query: Query[Any], values: Any, update_kwargs: Any
    ) -> None: ...

class BulkDelete(BulkUD): ...



from typing import Any
from typing import Optional

from .util import compat as compat

class SQLAlchemyError(Exception):
    code: Any = ...
    def __init__(self, *arg: Any, **kw: Any) -> None: ...
    def __unicode__(self): ...

class ArgumentError(SQLAlchemyError): ...

class ObjectNotExecutableError(ArgumentError):
    def __init__(self, target: Any) -> None: ...

class NoSuchModuleError(ArgumentError): ...
class NoForeignKeysError(ArgumentError): ...
class AmbiguousForeignKeysError(ArgumentError): ...

class CircularDependencyError(SQLAlchemyError):
    cycles: Any = ...
    edges: Any = ...
    def __init__(
        self,
        message: Any,
        cycles: Any,
        edges: Any,
        msg: Optional[Any] = ...,
        code: Optional[Any] = ...,
    ) -> None: ...
    def __reduce__(self): ...

class CompileError(SQLAlchemyError): ...

class UnsupportedCompilationError(CompileError):
    code: str = ...
    def __init__(
        self, compiler: Any, element_type: Any, message: Optional[Any] = ...
    ) -> None: ...

class IdentifierError(SQLAlchemyError): ...

class DisconnectionError(SQLAlchemyError):
    invalidate_pool: bool = ...

class InvalidatePoolError(DisconnectionError):
    invalidate_pool: bool = ...

class TimeoutError(SQLAlchemyError): ...
class InvalidRequestError(SQLAlchemyError): ...
class NoInspectionAvailable(InvalidRequestError): ...
class PendingRollbackError(InvalidRequestError): ...
class ResourceClosedError(InvalidRequestError): ...
class NoSuchColumnError(KeyError, InvalidRequestError): ...
class NoResultFound(InvalidRequestError): ...
class MultipleResultsFound(InvalidRequestError): ...
class NoReferenceError(InvalidRequestError): ...

class AwaitRequired(InvalidRequestError):
    code: str = ...

class MissingGreenlet(InvalidRequestError):
    code: str = ...

class NoReferencedTableError(NoReferenceError):
    table_name: Any = ...
    def __init__(self, message: Any, tname: Any) -> None: ...
    def __reduce__(self): ...

class NoReferencedColumnError(NoReferenceError):
    table_name: Any = ...
    column_name: Any = ...
    def __init__(self, message: Any, tname: Any, cname: Any) -> None: ...
    def __reduce__(self): ...

class NoSuchTableError(InvalidRequestError): ...
class UnreflectableTableError(InvalidRequestError): ...
class UnboundExecutionError(InvalidRequestError): ...
class DontWrapMixin: ...

class StatementError(SQLAlchemyError):
    statement: Any = ...
    params: Any = ...
    orig: Any = ...
    ismulti: Any = ...
    hide_parameters: Any = ...
    detail: Any = ...
    def __init__(
        self,
        message: Any,
        statement: Any,
        params: Any,
        orig: Any,
        hide_parameters: bool = ...,
        code: Optional[Any] = ...,
        ismulti: Optional[Any] = ...,
    ) -> None: ...
    def add_detail(self, msg: Any) -> None: ...
    def __reduce__(self): ...

class DBAPIError(StatementError):
    code: str = ...
    @classmethod
    def instance(
        cls,
        statement: Any,
        params: Any,
        orig: Any,
        dbapi_base_err: Any,
        hide_parameters: bool = ...,
        connection_invalidated: bool = ...,
        dialect: Optional[Any] = ...,
        ismulti: Optional[Any] = ...,
    ): ...
    def __reduce__(self): ...
    connection_invalidated: Any = ...
    def __init__(
        self,
        statement: Any,
        params: Any,
        orig: Any,
        hide_parameters: bool = ...,
        connection_invalidated: bool = ...,
        code: Optional[Any] = ...,
        ismulti: Optional[Any] = ...,
    ) -> None: ...

class InterfaceError(DBAPIError):
    code: str = ...

class DatabaseError(DBAPIError):
    code: str = ...

class DataError(DatabaseError):
    code: str = ...

class OperationalError(DatabaseError):
    code: str = ...

class IntegrityError(DatabaseError):
    code: str = ...

class InternalError(DatabaseError):
    code: str = ...

class ProgrammingError(DatabaseError):
    code: str = ...

class NotSupportedError(DatabaseError):
    code: str = ...

class SADeprecationWarning(DeprecationWarning):
    deprecated_since: Any = ...

class RemovedIn20Warning(SADeprecationWarning):
    deprecated_since: str = ...

class MovedIn20Warning(RemovedIn20Warning): ...

class SAPendingDeprecationWarning(PendingDeprecationWarning):
    deprecated_since: Any = ...

class SAWarning(RuntimeWarning): ...



from typing import Any
from typing import Collection
from typing import Generic
from typing import ItemsView
from typing import Iterator
from typing import KeysView
from typing import List
from typing import Mapping
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import ValuesView

from .result import ResultMetaData
from .result import RMKeyView

_T = TypeVar("_T")

MD_INDEX: int

def rowproxy_reconstructor(cls: Type[_T], state: Any) -> _T: ...

KEY_INTEGER_ONLY: int
KEY_OBJECTS_ONLY: int
KEY_OBJECTS_BUT_WARN: int
KEY_OBJECTS_NO_WARN: int

class BaseRow:
    def __init__(
        self,
        parent: ResultMetaData,
        processors: Any,
        keymap: Any,
        key_style: Any,
        data: Any,
    ) -> None: ...
    def __reduce__(self) -> Any: ...
    def __iter__(self) -> Iterator[Any]: ...
    def __len__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __getitem__(self, key: Any) -> Any: ...
    def __getattr__(self, name: Any) -> Any: ...

class Row(BaseRow, Sequence[Any]):
    @property
    def _mapping(self) -> RowMapping: ...
    def __contains__(self, key: Any) -> bool: ...
    def __hash__(self) -> int: ...
    def __lt__(self, other: Any) -> bool: ...
    def __le__(self, other: Any) -> bool: ...
    def __ge__(self, other: Any) -> bool: ...
    def __gt__(self, other: Any) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...
    def keys(self) -> List[str]: ...

class LegacyRow(Row):
    def __contains__(self, key: Any) -> bool: ...
    def __getitem__(self, key: Any) -> Any: ...
    def has_key(self, key: str) -> bool: ...
    def items(self) -> List[Tuple[str, Any]]: ...
    def iterkeys(self) -> Iterator[str]: ...
    def itervalues(self) -> Iterator[Any]: ...
    def values(self) -> List[Any]: ...

BaseRowProxy = BaseRow
RowProxy = Row

class ROMappingView(  # type: ignore[misc]
    KeysView[str],
    ValuesView[Any],
    ItemsView[str, Any],
    Generic[_T],
):
    def __init__(self, mapping: RowMapping, items: Collection[_T]) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[Tuple[str, _T]]: ...  # type: ignore[override]
    def __contains__(self, item: Any) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...

class RowMapping(BaseRow, Mapping[str, Any]):
    def __getitem__(self, key: str) -> Any: ...
    def __iter__(self) -> Iterator[str]: ...
    def __len__(self) -> int: ...
    def __contains__(self, key: Any) -> bool: ...
    def items(self) -> ROMappingView[Tuple[str, Any]]: ...  # type: ignore[override]
    def keys(self) -> RMKeyView: ...
    def values(self) -> ROMappingView[Any]: ...



from typing import Any
from typing import Callable
from typing import ContextManager
from typing import Iterable
from typing import Iterator
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import TypeVar
from typing import Union

from .base import StartableContext
from .engine import AsyncConnection
from .engine import AsyncEngine
from .result import AsyncResult
from ... import util
from ...engine import Result
from ...engine.base import _ExecutionOptions
from ...orm import Session
from ...orm.session import _IdentityMap
from ...sql import ClauseElement
from ...sql import Executable

_T = TypeVar("_T")
_TAsyncSession = TypeVar("_TAsyncSession", bound=AsyncSession)
_TAsyncSessionTransaction = TypeVar(
    "_TAsyncSessionTransaction", bound=AsyncSessionTransaction
)

class AsyncSession:
    dispatch: Any = ...
    bind: Any = ...
    binds: Any = ...
    sync_session: Session = ...
    def __init__(
        self,
        bind: Optional[Union[AsyncConnection, AsyncEngine]] = ...,
        binds: Optional[
            Mapping[object, Union[AsyncConnection, AsyncEngine]]
        ] = ...,
        **kw: Any,
    ) -> None: ...
    async def refresh(
        self,
        instance: Any,
        attribute_names: Optional[Any] = ...,
        with_for_update: Optional[Any] = ...,
    ) -> None: ...
    async def run_sync(
        self, fn: Callable[..., _T], *arg: Any, **kw: Any
    ) -> _T: ...
    async def execute(
        self,
        statement: Executable,
        params: Optional[
            Union[Mapping[str, Any], Sequence[Mapping[str, Any]]]
        ] = ...,
        execution_options: Optional[_ExecutionOptions] = ...,
        bind_arguments: Optional[Mapping[str, Any]] = ...,
        **kw: Any,
    ) -> Result: ...
    async def scalar(
        self,
        statement: Executable,
        params: Optional[
            Union[Mapping[str, Any], Sequence[Mapping[str, Any]]]
        ] = ...,
        execution_options: _ExecutionOptions = ...,
        bind_arguments: Optional[Mapping[str, Any]] = ...,
        **kw: Any,
    ) -> Any: ...
    async def get(
        self,
        entity: Any,
        ident: Any,
        options: Optional[Any] = ...,
        populate_existing: bool = ...,
        with_for_update: Optional[Any] = ...,
        identity_token: Optional[Any] = ...,
    ) -> Any: ...
    async def stream(
        self,
        statement: Any,
        params: Optional[
            Union[Mapping[str, Any], Sequence[Mapping[str, Any]]]
        ] = ...,
        execution_options: _ExecutionOptions = ...,
        bind_arguments: Optional[Mapping[str, Any]] = ...,
        **kw: Any,
    ) -> AsyncResult: ...
    async def delete(self, instance: Any) -> None: ...
    async def merge(self, instance: _T, load: bool = ...) -> _T: ...
    async def flush(self, objects: Optional[Any] = ...) -> None: ...
    async def connection(self) -> AsyncConnection: ...
    def begin(self, **kw: Any) -> AsyncSessionTransaction: ...
    def begin_nested(self, **kw: Any) -> AsyncSessionTransaction: ...
    async def rollback(self) -> None: ...
    async def commit(self) -> None: ...
    async def close(self) -> None: ...
    @classmethod
    async def close_all(self) -> None: ...
    async def __aenter__(self: _TAsyncSession) -> _TAsyncSession: ...
    async def __aexit__(
        self, type_: Any, value: Any, traceback: Any
    ) -> None: ...
    # copied via create_proxy_methods
    def __contains__(self, instance: Any) -> bool: ...
    def __iter__(self) -> Iterator[Any]: ...
    def add(self, instance: Any, _warn: bool = ...) -> None: ...
    def add_all(self, instances: Any) -> None: ...
    def expire(
        self, instance: Any, attribute_names: Optional[Iterable[str]] = ...
    ) -> None: ...
    def expire_all(self) -> None: ...
    def expunge(self, instance: Any) -> None: ...
    def expunge_all(self) -> None: ...
    def get_bind(
        self,
        mapper: Optional[Any] = ...,
        clause: Optional[ClauseElement] = ...,
        bind: Optional[Union[AsyncConnection, AsyncEngine]] = ...,
        _sa_skip_events: Optional[Any] = ...,
        _sa_skip_for_implicit_returning: bool = ...,
    ) -> Union[AsyncConnection, AsyncEngine]: ...
    def is_modified(
        self, instance: Any, include_collections: bool = ...
    ) -> bool: ...
    def in_transaction(self) -> bool: ...
    @property
    def dirty(self) -> util.IdentitySet[Any]: ...
    @property
    def deleted(self) -> util.IdentitySet[Any]: ...
    @property
    def new(self) -> util.IdentitySet[Any]: ...
    identity_map: _IdentityMap
    @property
    def is_active(self) -> bool: ...
    autoflush: bool
    @property
    def no_autoflush(
        self: _TAsyncSession,
    ) -> ContextManager[_TAsyncSession]: ...
    @util.memoized_property
    def info(self) -> Mapping[Any, Any]: ...

class _AsyncSessionContextManager:
    async_session: AsyncSession = ...
    trans: AsyncSessionTransaction = ...
    def __init__(self, async_session: AsyncSession) -> None: ...
    async def __aenter__(self) -> AsyncSession: ...
    async def __aexit__(
        self, type_: Any, value: Any, traceback: Any
    ) -> None: ...

class AsyncSessionTransaction(StartableContext):
    session: AsyncSession = ...
    nested: bool = ...
    sync_transaction: Optional[Any] = ...
    def __init__(self, session: AsyncSession, nested: bool = ...) -> None: ...
    @property
    def is_active(self) -> bool: ...
    async def rollback(self) -> Optional[AsyncSessionTransaction]: ...
    async def commit(self) -> Optional[AsyncSessionTransaction]: ...
    async def start(
        self: _TAsyncSessionTransaction,
    ) -> _TAsyncSessionTransaction: ...
    async def __aexit__(
        self, type_: Any, value: Any, traceback: Any
    ) -> None: ...



import sys
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Pattern
from typing import Set
from typing import Tuple
from typing import TypeVar
from typing import Union

from . import dml
from . import elements
from . import selectable
from .base import CompileState
from .elements import ClauseElement
from .schema import Table
from .type_api import TypeEngine
from .. import util
from ..engine import Dialect
from ..util import langhelpers

_T = TypeVar("_T")
_U = TypeVar("_U")
_TE = TypeVar("_TE", bound=TypeEngine[Any])
_CL = TypeVar("_CL", bound=_CompileLabel[Any])
_SC = TypeVar("_SC", bound=SQLCompiler)

if sys.version_info >= (3, 0):
    _VisitResult = str
else:
    _VisitResult = Union[str, unicode]  # noqa

RESERVED_WORDS: Set[str]
LEGAL_CHARACTERS: Pattern[str]
LEGAL_CHARACTERS_PLUS_SPACE: Pattern[str]
ILLEGAL_INITIAL_CHARACTERS: str
FK_ON_DELETE: Pattern[str]
FK_ON_UPDATE: Pattern[str]
FK_INITIALLY: Pattern[str]
BIND_PARAMS: Pattern[str]
BIND_PARAMS_ESC: Pattern[str]
BIND_TEMPLATES: Dict[str, str]
BIND_TRANSLATE: Dict[str, Pattern[str]]
OPERATORS: Dict[Any, str]
FUNCTIONS: Dict[Any, str]
EXTRACT_MAP: Dict[str, str]
COMPOUND_KEYWORDS: Dict[langhelpers._symbol, str]
RM_RENDERED_NAME: int
RM_NAME: int
RM_OBJECTS: int
RM_TYPE: int

class ExpandedState(NamedTuple):
    statement: Any
    additional_parameters: Any
    processors: Any
    positiontup: Any
    parameter_expansion: Any

NO_LINTING: langhelpers._symbol
COLLECT_CARTESIAN_PRODUCTS: langhelpers._symbol
WARN_LINTING: langhelpers._symbol
FROM_LINTING: langhelpers._symbol

class FromLinter(NamedTuple):
    froms: Dict[Table, str]
    edges: Set[Tuple[Table, Table]]
    def lint(
        self, start: Optional[Table] = ...
    ) -> Union[Tuple[Set[Table], Table], Tuple[None, None]]: ...
    def warn(self) -> None: ...

class Compiled:
    schema_translate_map: Optional[Dict[Any, Any]] = ...
    execution_options: util.immutabledict[Any, Any] = ...
    compile_state: Optional[CompileState] = ...
    cache_key: Any = ...
    dialect: Dialect = ...
    preparer: Any = ...
    statement: Optional[ClauseElement] = ...
    can_execute: bool = ...
    string: util.text_type = ...
    def __init__(
        self,
        dialect: Dialect,
        statement: Optional[Dict[Any, Any]],
        schema_translate_map: Optional[Any] = ...,
        render_schema_translate: bool = ...,
        compile_kwargs: Any = ...,
    ) -> None: ...
    def visit_unsupported_compilation(
        self, element: Any, err: Any
    ) -> _VisitResult: ...
    @property
    def sql_compiler(self) -> Compiled: ...
    def process(self, obj: Any, **kwargs: Any) -> _VisitResult: ...
    def construct_params(
        self,
        params: Optional[Any] = ...,
        extracted_parameters: Optional[Any] = ...,
    ) -> Optional[Dict[str, Any]]: ...
    @property
    def params(self) -> Optional[Dict[str, Any]]: ...

class TypeCompiler(object, metaclass=util.EnsureKWArgType):
    ensure_kwarg: str = ...
    dialect: Dialect = ...
    def __init__(self, dialect: Dialect) -> None: ...
    def process(self, type_: TypeEngine[Any], **kw: Any) -> _VisitResult: ...
    def visit_unsupported_compilation(
        self, element: Any, err: Any, **kw: Any
    ) -> _VisitResult: ...

class _CompileLabel(elements.ColumnElement[_TE]):
    __visit_name__: str = ...
    element: elements.ColumnElement[_TE] = ...
    name: Any = ...
    def __init__(
        self, col: elements.ColumnElement[_TE], name: str, alt_names: Any = ...
    ) -> None: ...
    @property
    def proxy_set(self) -> util.column_set[Any]: ...  # type: ignore[override]
    @property
    def type(self) -> _TE: ...  # type: ignore[override]
    def self_group(self: _CL, **kw: Any) -> _CL: ...  # type: ignore[override]

class SQLCompiler(Compiled):
    extract_map: Any = ...
    compound_keywords: Any = ...
    isdelete: bool = ...
    isinsert: bool = ...
    isupdate: bool = ...
    isplaintext: bool = ...
    returning: Any = ...
    returning_precedes_values: bool = ...
    render_table_with_column_in_update_from: bool = ...
    ansi_bind_rules: bool = ...
    insert_single_values_expr: Any = ...
    literal_execute_params: Any = ...
    post_compile_params: Any = ...
    escaped_bind_names: Any = ...
    has_out_parameters: bool = ...
    insert_prefetch: Any = ...
    update_prefetch: Any = ...
    postfetch_lastrowid: bool = ...
    inline: bool = ...
    column_keys: Any = ...
    cache_key: Any = ...
    for_executemany: Any = ...
    linting: Any = ...
    binds: Any = ...
    bind_names: Any = ...
    stack: Any = ...
    positional: Any = ...
    positiontup: Any = ...
    bindtemplate: Any = ...
    ctes: Any = ...
    label_length: Any = ...
    anon_map: Any = ...
    truncated_names: Any = ...
    ctes_recursive: bool = ...
    translate_select_structure: Any = ...
    def __init__(
        self,
        dialect: Dialect,
        statement: Any,
        cache_key: Optional[Any] = ...,
        column_keys: Optional[Any] = ...,
        for_executemany: bool = ...,
        linting: Any = ...,
        **kwargs: Any,
    ) -> None: ...
    @property
    def current_executable(
        self,
    ) -> Union[
        selectable.Select,
        dml.Insert,
        dml.Update,
        dml.Delete,
        selectable.CompoundSelect,
    ]: ...
    @property
    def prefetch(self) -> List[Any]: ...
    def is_subquery(self) -> bool: ...
    @property
    def sql_compiler(self: _SC) -> _SC: ...
    def construct_params(  # type: ignore[override]
        self,
        params: Optional[Any] = ...,
        _group_number: Optional[Any] = ...,
        _check: bool = ...,
        extracted_parameters: Optional[Any] = ...,
    ) -> Optional[Dict[str, Any]]: ...
    @property
    def params(self) -> Dict[str, Any]: ...
    def default_from(self) -> str: ...
    def visit_grouping(
        self, grouping: Any, asfrom: bool = ..., **kwargs: Any
    ) -> _VisitResult: ...
    def visit_label_reference(
        self, element: Any, within_columns_clause: bool = ..., **kwargs: Any
    ) -> _VisitResult: ...
    def visit_textual_label_reference(
        self, element: Any, within_columns_clause: bool = ..., **kwargs: Any
    ) -> _VisitResult: ...
    def visit_label(
        self,
        label: Any,
        add_to_result_map: Optional[Any] = ...,
        within_label_clause: bool = ...,
        within_columns_clause: bool = ...,
        render_label_as_label: Optional[Any] = ...,
        result_map_targets: Any = ...,
        **kw: Any,
    ) -> _VisitResult: ...
    def visit_lambda_element(
        self, element: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_column(
        self,
        column: Any,
        add_to_result_map: Optional[Any] = ...,
        include_table: bool = ...,
        result_map_targets: Any = ...,
        **kwargs: Any,
    ) -> _VisitResult: ...
    def visit_collation(self, element: Any, **kw: Any) -> _VisitResult: ...
    def visit_fromclause(
        self, fromclause: Any, **kwargs: Any
    ) -> _VisitResult: ...
    def visit_index(self, index: Any, **kwargs: Any) -> _VisitResult: ...
    def visit_typeclause(self, typeclause: Any, **kw: Any) -> _VisitResult: ...
    def post_process_text(self, text: Any) -> _VisitResult: ...
    def escape_literal_column(self, text: Any) -> _VisitResult: ...
    def visit_textclause(
        self,
        textclause: Any,
        add_to_result_map: Optional[Any] = ...,
        **kw: Any,
    ) -> _VisitResult: ...
    def visit_textual_select(
        self,
        taf: Any,
        compound_index: Optional[Any] = ...,
        asfrom: bool = ...,
        **kw: Any,
    ) -> _VisitResult: ...
    def visit_null(self, expr: Any, **kw: Any) -> _VisitResult: ...
    def visit_true(self, expr: Any, **kw: Any) -> _VisitResult: ...
    def visit_false(self, expr: Any, **kw: Any) -> _VisitResult: ...
    def visit_tuple(self, clauselist: Any, **kw: Any) -> _VisitResult: ...
    def visit_clauselist(self, clauselist: Any, **kw: Any) -> _VisitResult: ...
    def visit_case(self, clause: Any, **kwargs: Any) -> _VisitResult: ...
    def visit_type_coerce(
        self, type_coerce: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_cast(self, cast: Any, **kwargs: Any) -> _VisitResult: ...
    def visit_over(self, over: Any, **kwargs: Any) -> _VisitResult: ...
    def visit_withingroup(
        self, withingroup: Any, **kwargs: Any
    ) -> _VisitResult: ...
    def visit_funcfilter(
        self, funcfilter: Any, **kwargs: Any
    ) -> _VisitResult: ...
    def visit_extract(self, extract: Any, **kwargs: Any) -> _VisitResult: ...
    def visit_scalar_function_column(
        self, element: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_function(
        self, func: Any, add_to_result_map: Optional[Any] = ..., **kwargs: Any
    ) -> _VisitResult: ...
    def visit_next_value_func(
        self, next_value: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_sequence(self, sequence: Any, **kw: Any) -> _VisitResult: ...
    def function_argspec(self, func: Any, **kwargs: Any) -> _VisitResult: ...
    def visit_compound_select(
        self,
        cs: Any,
        asfrom: bool = ...,
        compound_index: Optional[Any] = ...,
        **kwargs: Any,
    ) -> _VisitResult: ...
    def visit_unary(self, unary: Any, **kw: Any) -> _VisitResult: ...
    def visit_is_true_unary_operator(
        self, element: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_is_false_unary_operator(
        self, element: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_not_match_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_empty_set_expr(self, element_types: Any) -> _VisitResult: ...
    def visit_binary(
        self,
        binary: Any,
        override_operator: Optional[Any] = ...,
        eager_grouping: bool = ...,
        from_linter: Optional[Any] = ...,
        lateral_from_linter: Optional[Any] = ...,
        **kw: Any,
    ) -> _VisitResult: ...
    def visit_function_as_comparison_op_binary(
        self, element: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_mod_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_custom_op_binary(
        self, element: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_custom_op_unary_operator(
        self, element: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_custom_op_unary_modifier(
        self, element: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_contains_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_not_contains_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_startswith_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_not_startswith_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_endswith_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_not_endswith_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_like_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_not_like_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_ilike_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_not_ilike_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_between_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_not_between_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_regexp_match_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_not_regexp_match_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_regexp_replace_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_bindparam(
        self,
        bindparam: Any,
        within_columns_clause: bool = ...,
        literal_binds: bool = ...,
        skip_bind_expression: bool = ...,
        literal_execute: bool = ...,
        render_postcompile: bool = ...,
        **kwargs: Any,
    ) -> _VisitResult: ...
    def render_literal_bindparam(
        self, bindparam: Any, render_literal_value: Any = ..., **kw: Any
    ) -> _VisitResult: ...
    def render_literal_value(self, value: Any, type_: Any) -> _VisitResult: ...
    def bindparam_string(
        self,
        name: Any,
        positional_names: Optional[Any] = ...,
        post_compile: bool = ...,
        expanding: bool = ...,
        escaped_from: Optional[Any] = ...,
        **kw: Any,
    ) -> _VisitResult: ...
    def visit_cte(
        self,
        cte: Any,
        asfrom: bool = ...,
        ashint: bool = ...,
        fromhints: Optional[Any] = ...,
        visiting_cte: Optional[Any] = ...,
        from_linter: Optional[Any] = ...,
        **kwargs: Any,
    ) -> _VisitResult: ...
    def visit_table_valued_alias(
        self, element: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_table_valued_column(
        self, element: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_alias(
        self,
        alias: Any,
        asfrom: bool = ...,
        ashint: bool = ...,
        iscrud: bool = ...,
        fromhints: Optional[Any] = ...,
        subquery: bool = ...,
        lateral: bool = ...,
        enclosing_alias: Optional[Any] = ...,
        from_linter: Optional[Any] = ...,
        **kwargs: Any,
    ) -> _VisitResult: ...
    def visit_subquery(self, subquery: Any, **kw: Any) -> _VisitResult: ...
    def visit_lateral(self, lateral_: Any, **kw: Any) -> _VisitResult: ...
    def visit_tablesample(
        self, tablesample: Any, asfrom: bool = ..., **kw: Any
    ) -> _VisitResult: ...
    def visit_values(
        self,
        element: Any,
        asfrom: bool = ...,
        from_linter: Optional[Any] = ...,
        **kw: Any,
    ) -> _VisitResult: ...
    def get_render_as_alias_suffix(
        self, alias_name_text: Any
    ) -> _VisitResult: ...
    def format_from_hint_text(
        self, sqltext: Any, table: Any, hint: Any, iscrud: Any
    ) -> _VisitResult: ...
    def get_select_hint_text(self, byfroms: Any) -> Optional[str]: ...
    def get_from_hint_text(self, table: Any, text: Any) -> Optional[str]: ...
    def get_crud_hint_text(self, table: Any, text: Any) -> Optional[str]: ...
    def get_statement_hint_text(self, hint_texts: Any) -> _VisitResult: ...
    def visit_select(
        self,
        select_stmt: Any,
        asfrom: bool = ...,
        fromhints: Optional[Any] = ...,
        compound_index: Optional[Any] = ...,
        select_wraps_for: Optional[Any] = ...,
        lateral: bool = ...,
        from_linter: Optional[Any] = ...,
        **kwargs: Any,
    ) -> _VisitResult: ...
    def get_cte_preamble(self, recursive: Any) -> _VisitResult: ...
    def get_select_precolumns(
        self, select: Any, **kw: Any
    ) -> _VisitResult: ...
    def group_by_clause(self, select: Any, **kw: Any) -> _VisitResult: ...
    def order_by_clause(self, select: Any, **kw: Any) -> _VisitResult: ...
    def for_update_clause(self, select: Any, **kw: Any) -> _VisitResult: ...
    def returning_clause(
        self, stmt: Any, returning_cols: Any
    ) -> _VisitResult: ...
    def limit_clause(self, select: Any, **kw: Any) -> _VisitResult: ...
    def fetch_clause(self, select: Any, **kw: Any) -> _VisitResult: ...
    def visit_table(
        self,
        table: Any,
        asfrom: bool = ...,
        iscrud: bool = ...,
        ashint: bool = ...,
        fromhints: Optional[Any] = ...,
        use_schema: bool = ...,
        from_linter: Optional[Any] = ...,
        **kwargs: Any,
    ) -> _VisitResult: ...
    def visit_join(
        self,
        join: Any,
        asfrom: bool = ...,
        from_linter: Optional[Any] = ...,
        **kwargs: Any,
    ) -> _VisitResult: ...
    def visit_insert(self, insert_stmt: Any, **kw: Any) -> _VisitResult: ...
    def update_limit_clause(self, update_stmt: Any) -> _VisitResult: ...
    def update_tables_clause(
        self, update_stmt: Any, from_table: Any, extra_froms: Any, **kw: Any
    ) -> _VisitResult: ...
    def update_from_clause(
        self,
        update_stmt: Any,
        from_table: Any,
        extra_froms: Any,
        from_hints: Any,
        **kw: Any,
    ) -> _VisitResult: ...
    def visit_update(self, update_stmt: Any, **kw: Any) -> _VisitResult: ...
    def delete_extra_from_clause(
        self,
        update_stmt: Any,
        from_table: Any,
        extra_froms: Any,
        from_hints: Any,
        **kw: Any,
    ) -> _VisitResult: ...
    def delete_table_clause(
        self, delete_stmt: Any, from_table: Any, extra_froms: Any
    ) -> _VisitResult: ...
    def visit_delete(self, delete_stmt: Any, **kw: Any) -> _VisitResult: ...
    def visit_savepoint(self, savepoint_stmt: Any) -> _VisitResult: ...
    def visit_rollback_to_savepoint(
        self, savepoint_stmt: Any
    ) -> _VisitResult: ...
    def visit_release_savepoint(self, savepoint_stmt: Any) -> _VisitResult: ...

class StrSQLCompiler(SQLCompiler):
    def visit_unsupported_compilation(
        self, element: Any, err: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_getitem_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_json_getitem_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_json_path_getitem_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_sequence(self, seq: Any, **kw: Any) -> _VisitResult: ...  # type: ignore[override]
    def returning_clause(
        self, stmt: Any, returning_cols: Any
    ) -> _VisitResult: ...
    def update_from_clause(
        self,
        update_stmt: Any,
        from_table: Any,
        extra_froms: Any,
        from_hints: Any,
        **kw: Any,
    ) -> _VisitResult: ...
    def delete_extra_from_clause(
        self,
        update_stmt: Any,
        from_table: Any,
        extra_froms: Any,
        from_hints: Any,
        **kw: Any,
    ) -> _VisitResult: ...
    def visit_empty_set_expr(self, type_: Any) -> _VisitResult: ...
    def get_from_hint_text(self, table: Any, text: Any) -> _VisitResult: ...
    def visit_regexp_match_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_not_regexp_match_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_regexp_replace_op_binary(
        self, binary: Any, operator: Any, **kw: Any
    ) -> _VisitResult: ...

class DDLCompiler(Compiled):
    @util.memoized_property
    def type_compiler(self) -> GenericTypeCompiler: ...
    def construct_params(
        self,
        params: Optional[Any] = ...,
        extracted_parameters: Optional[Any] = ...,
    ) -> Optional[Dict[str, Any]]: ...
    def visit_ddl(self, ddl: Any, **kwargs: Any) -> _VisitResult: ...
    def visit_create_schema(self, create: Any, **kw: Any) -> _VisitResult: ...
    def visit_drop_schema(self, drop: Any, **kw: Any) -> _VisitResult: ...
    def visit_create_table(self, create: Any, **kw: Any) -> _VisitResult: ...
    def visit_create_column(
        self, create: Any, first_pk: bool = ..., **kw: Any
    ) -> _VisitResult: ...
    def create_table_constraints(
        self,
        table: Any,
        _include_foreign_key_constraints: Optional[Any] = ...,
        **kw: Any,
    ) -> _VisitResult: ...
    def visit_drop_table(self, drop: Any, **kw: Any) -> _VisitResult: ...
    def visit_drop_view(self, drop: Any, **kw: Any) -> _VisitResult: ...
    def visit_create_index(
        self,
        create: Any,
        include_schema: bool = ...,
        include_table_schema: bool = ...,
        **kw: Any,
    ) -> _VisitResult: ...
    def visit_drop_index(self, drop: Any, **kw: Any) -> _VisitResult: ...
    def visit_add_constraint(self, create: Any, **kw: Any) -> _VisitResult: ...
    def visit_set_table_comment(
        self, create: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_drop_table_comment(
        self, drop: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_set_column_comment(
        self, create: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_drop_column_comment(
        self, drop: Any, **kw: Any
    ) -> _VisitResult: ...
    def get_identity_options(self, identity_options: Any) -> _VisitResult: ...
    def visit_create_sequence(
        self, create: Any, prefix: Optional[Any] = ..., **kw: Any
    ) -> _VisitResult: ...
    def visit_drop_sequence(self, drop: Any, **kw: Any) -> _VisitResult: ...
    def visit_drop_constraint(self, drop: Any, **kw: Any) -> _VisitResult: ...
    def get_column_specification(
        self, column: Any, **kwargs: Any
    ) -> _VisitResult: ...
    def create_table_suffix(self, table: Any) -> _VisitResult: ...
    def post_create_table(self, table: Any) -> _VisitResult: ...
    def get_column_default_string(self, column: Any) -> _VisitResult: ...
    def visit_table_or_column_check_constraint(
        self, constraint: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_check_constraint(
        self, constraint: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_column_check_constraint(
        self, constraint: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_primary_key_constraint(
        self, constraint: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_foreign_key_constraint(
        self, constraint: Any, **kw: Any
    ) -> _VisitResult: ...
    def define_constraint_remote_table(
        self, constraint: Any, table: Any, preparer: Any
    ) -> _VisitResult: ...
    def visit_unique_constraint(
        self, constraint: Any, **kw: Any
    ) -> _VisitResult: ...
    def define_constraint_cascades(self, constraint: Any) -> _VisitResult: ...
    def define_constraint_deferrability(
        self, constraint: Any
    ) -> _VisitResult: ...
    def define_constraint_match(self, constraint: Any) -> _VisitResult: ...
    def visit_computed_column(
        self, generated: Any, **kw: Any
    ) -> _VisitResult: ...
    def visit_identity_column(
        self, identity: Any, **kw: Any
    ) -> _VisitResult: ...

class GenericTypeCompiler(TypeCompiler):
    def visit_FLOAT(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_REAL(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_NUMERIC(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_DECIMAL(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_INTEGER(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_SMALLINT(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_BIGINT(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_TIMESTAMP(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_DATETIME(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_DATE(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_TIME(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_CLOB(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_NCLOB(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_CHAR(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_NCHAR(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_VARCHAR(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_NVARCHAR(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_TEXT(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_BLOB(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_BINARY(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_VARBINARY(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_BOOLEAN(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_large_binary(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_boolean(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_time(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_datetime(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_date(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_big_integer(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_small_integer(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_integer(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_real(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_float(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_numeric(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_string(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_unicode(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_text(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_unicode_text(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_enum(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_null(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_type_decorator(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_user_defined(self, type_: Any, **kw: Any) -> _VisitResult: ...

class StrSQLTypeCompiler(GenericTypeCompiler):
    def process(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def __getattr__(self, key: Any) -> Any: ...
    def visit_null(self, type_: Any, **kw: Any) -> _VisitResult: ...
    def visit_user_defined(self, type_: Any, **kw: Any) -> _VisitResult: ...

class IdentifierPreparer:
    reserved_words: Any = ...
    legal_characters: Any = ...
    illegal_initial_characters: Any = ...
    schema_for_object: Any = ...
    dialect: Dialect = ...
    initial_quote: Any = ...
    final_quote: Any = ...
    escape_quote: Any = ...
    escape_to_quote: Any = ...
    omit_schema: Any = ...
    quote_case_sensitive_collations: Any = ...
    def __init__(
        self,
        dialect: Dialect,
        initial_quote: str = ...,
        final_quote: Optional[Any] = ...,
        escape_quote: str = ...,
        quote_case_sensitive_collations: bool = ...,
        omit_schema: bool = ...,
    ) -> None: ...
    def validate_sql_phrase(self, element: _T, reg: Any) -> _T: ...
    def quote_identifier(self, value: Any) -> _VisitResult: ...
    def quote_schema(
        self, schema: Any, force: Optional[Any] = ...
    ) -> _VisitResult: ...
    def quote(
        self, ident: Any, force: Optional[Any] = ...
    ) -> _VisitResult: ...
    def format_collation(self, collation_name: Any) -> _VisitResult: ...
    def format_sequence(
        self, sequence: Any, use_schema: bool = ...
    ) -> _VisitResult: ...
    def format_label(
        self, label: Any, name: Optional[Any] = ...
    ) -> _VisitResult: ...
    def format_alias(
        self, alias: Any, name: Optional[Any] = ...
    ) -> _VisitResult: ...
    def format_savepoint(
        self, savepoint: Any, name: Optional[Any] = ...
    ) -> _VisitResult: ...
    def format_constraint(
        self, constraint: Any, _alembic_quote: bool = ...
    ) -> _VisitResult: ...
    def format_index(self, index: Any) -> _VisitResult: ...
    def format_table(
        self, table: Any, use_schema: bool = ..., name: Optional[Any] = ...
    ) -> _VisitResult: ...
    def format_schema(self, name: Any) -> _VisitResult: ...
    def format_column(
        self,
        column: Any,
        use_table: bool = ...,
        name: Optional[Any] = ...,
        table_name: Optional[Any] = ...,
        use_schema: bool = ...,
    ) -> _VisitResult: ...
    def format_table_seq(
        self, table: Any, use_schema: bool = ...
    ) -> _VisitResult: ...
    def unformat_identifiers(self, identifiers: Any) -> _VisitResult: ...



# fmt: off
from typing import Any
from typing import Optional

from .. import util as util
from ..orm import attributes as attributes
from ..orm import collections as collections
from ..orm.instrumentation import ClassManager as ClassManager
from ..orm.instrumentation import InstrumentationFactory as InstrumentationFactory
# fmt: on

INSTRUMENTATION_MANAGER: str

def find_native_user_instrumentation_hook(cls): ...

instrumentation_finders: Any

class ExtendedInstrumentationRegistry(InstrumentationFactory):
    def unregister(self, class_: Any) -> None: ...
    def manager_of_class(self, cls: Any): ...
    def state_of(self, instance: Any): ...
    def dict_of(self, instance: Any): ...

class InstrumentationManager:
    def __init__(self, class_: Any) -> None: ...
    def manage(self, class_: Any, manager: Any) -> None: ...
    def dispose(self, class_: Any, manager: Any) -> None: ...
    def manager_getter(self, class_: Any): ...
    def instrument_attribute(
        self, class_: Any, key: Any, inst: Any
    ) -> None: ...
    def post_configure_attribute(
        self, class_: Any, key: Any, inst: Any
    ) -> None: ...
    def install_descriptor(self, class_: Any, key: Any, inst: Any) -> None: ...
    def uninstall_descriptor(self, class_: Any, key: Any) -> None: ...
    def install_member(
        self, class_: Any, key: Any, implementation: Any
    ) -> None: ...
    def uninstall_member(self, class_: Any, key: Any) -> None: ...
    def instrument_collection_class(
        self, class_: Any, key: Any, collection_class: Any
    ): ...
    def get_instance_dict(self, class_: Any, instance: Any): ...
    def initialize_instance_dict(self, class_: Any, instance: Any) -> None: ...
    def install_state(
        self, class_: Any, instance: Any, state: Any
    ) -> None: ...
    def remove_state(self, class_: Any, instance: Any) -> None: ...
    def state_getter(self, class_: Any): ...
    def dict_getter(self, class_: Any): ...

class _ClassInstrumentationAdapter(ClassManager):
    def __init__(self, class_: Any, override: Any) -> None: ...
    def manage(self) -> None: ...
    def dispose(self) -> None: ...
    def manager_getter(self): ...
    def instrument_attribute(
        self, key: Any, inst: Any, propagated: bool = ...
    ) -> None: ...
    def post_configure_attribute(self, key: Any) -> None: ...
    def install_descriptor(self, key: Any, inst: Any) -> None: ...
    def uninstall_descriptor(self, key: Any) -> None: ...
    def install_member(self, key: Any, implementation: Any) -> None: ...
    def uninstall_member(self, key: Any) -> None: ...
    def instrument_collection_class(self, key: Any, collection_class: Any): ...
    def initialize_collection(self, key: Any, state: Any, factory: Any): ...
    def new_instance(self, state: Optional[Any] = ...): ...
    def setup_instance(self, instance: Any, state: Optional[Any] = ...): ...
    def teardown_instance(self, instance: Any) -> None: ...
    def has_state(self, instance: Any): ...
    def state_getter(self): ...
    def dict_getter(self): ...



from typing import Any
from typing import Optional
from typing import TypeVar
from typing import Union

from . import roles
from .base import CompileState
from .base import DialectKWArgs
from .base import Executable
from .base import HasCompileState
from .base import ImmutableColumnCollection
from .elements import BooleanClauseList
from .elements import ClauseElement
from .schema import Table
from .selectable import HasCTE
from .selectable import HasPrefixes

_UB = TypeVar("_UB", bound=UpdateBase)
_VB = TypeVar("_VB", bound=ValuesBase)
_I = TypeVar("_I", bound=Insert)
_DWB = TypeVar("_DWB", bound=DMLWhereBase)
_UD = TypeVar("_UD", bound=Update)

class DMLState(CompileState):
    isupdate: bool = ...
    isdelete: bool = ...
    isinsert: bool = ...
    def __init__(self, statement: Any, compiler: Any, **kw: Any) -> None: ...
    @property
    def dml_table(self) -> Table: ...

class InsertDMLState(DMLState):
    isinsert: bool = ...
    include_table_with_column_exprs: bool = ...
    statement: Any = ...
    def __init__(self, statement: Any, compiler: Any, **kw: Any) -> None: ...

class UpdateDMLState(DMLState):
    isupdate: bool = ...
    include_table_with_column_exprs: bool = ...
    statement: Any = ...
    is_multitable: bool = ...
    def __init__(self, statement: Any, compiler: Any, **kw: Any) -> None: ...

class DeleteDMLState(DMLState):
    isdelete: bool = ...
    statement: Any = ...
    def __init__(self, statement: Any, compiler: Any, **kw: Any) -> None: ...

class UpdateBase(
    roles.DMLRole,
    HasCTE,
    HasCompileState,
    DialectKWArgs,
    HasPrefixes,
    Executable,
    ClauseElement,
):
    __visit_name__: str = ...
    named_with_column: bool = ...
    is_dml: bool = ...
    bind: Any = ...
    def params(self: _UB, *arg: Any, **kw: Any) -> _UB: ...
    def with_dialect_options(self: _UB, **opt: Any) -> _UB: ...
    def returning(self: _UB, *cols: Any) -> _UB: ...
    @property
    def exported_columns(self) -> ImmutableColumnCollection[Any]: ...
    def with_hint(
        self: _UB,
        text: Any,
        selectable: Optional[Any] = ...,
        dialect_name: str = ...,
    ) -> _UB: ...

class ValuesBase(UpdateBase):
    __visit_name__: str = ...
    select: Any = ...
    table: Any = ...
    def __init__(self, table: Any, values: Any, prefixes: Any) -> None: ...
    def values(self: _VB, *args: Any, **kwargs: Any) -> _VB: ...
    def return_defaults(self: _VB, *cols: Any) -> _VB: ...

class Insert(ValuesBase):
    __visit_name__: str = ...
    select: Any = ...
    include_insert_from_select_defaults: bool = ...
    is_insert: bool = ...
    def __init__(
        self,
        table: Any,
        values: Optional[Any] = ...,
        inline: bool = ...,
        bind: Optional[Any] = ...,
        prefixes: Optional[Any] = ...,
        returning: Optional[Any] = ...,
        return_defaults: bool = ...,
        **dialect_kw: Any,
    ) -> None: ...
    def inline(self: _I) -> _I: ...
    def from_select(
        self: _I, names: Any, select: Any, include_defaults: bool = ...
    ) -> _I: ...

class DMLWhereBase:
    def where(self: _DWB, *whereclause: Any) -> _DWB: ...
    def filter(self: _DWB, *criteria: Any) -> _DWB: ...
    def filter_by(self: _DWB, **kwargs: Any) -> _DWB: ...
    @property
    def whereclause(
        self,
    ) -> Optional[Union[BooleanClauseList[Any], ClauseElement]]: ...

class Update(DMLWhereBase, ValuesBase):
    __visit_name__: str = ...
    is_update: bool = ...
    def __init__(
        self,
        table: Any,
        whereclause: Optional[Any] = ...,
        values: Optional[Any] = ...,
        inline: bool = ...,
        bind: Optional[Any] = ...,
        prefixes: Optional[Any] = ...,
        returning: Optional[Any] = ...,
        return_defaults: bool = ...,
        preserve_parameter_order: bool = ...,
        **dialect_kw: Any,
    ) -> None: ...
    def ordered_values(self: _UD, *args: Any) -> _UD: ...
    def inline(self: _UD) -> _UD: ...

class Delete(DMLWhereBase, UpdateBase):
    __visit_name__: str = ...
    is_delete: bool = ...
    table: Any = ...
    def __init__(
        self,
        table: Any,
        whereclause: Optional[Any] = ...,
        bind: Optional[Any] = ...,
        returning: Optional[Any] = ...,
        prefixes: Optional[Any] = ...,
        **dialect_kw: Any,
    ) -> None: ...



from typing import Any
from typing import Callable
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sequence
from typing import TypeVar
from typing import Union

from typing_extensions import Protocol

from . import visitors
from .base import ColumnCollection
from .base import ColumnSet
from .elements import ClauseElement
from .elements import ColumnClause
from .schema import Table
from .selectable import Alias
from .selectable import FromClause
from .selectable import Join
from .selectable import Selectable

_T = TypeVar("_T")
_COLA = TypeVar("_COLA", bound=ColumnAdapter)

class _IsDerivedFrom(Protocol):
    def is_derived_from(self, __fromclause: FromClause) -> bool: ...

join_condition = Join._join_condition

def find_join_source(
    clauses: Iterable[_IsDerivedFrom], join_to: Selectable
) -> List[int]: ...
def find_left_clause_that_matches_given(
    clauses: Sequence[_IsDerivedFrom], join_from: Selectable
) -> List[int]: ...
def find_left_clause_to_join_from(
    clauses: Sequence[FromClause],
    join_to: Selectable,
    onclause: Optional[Any],
) -> Iterable[int]: ...
def visit_binary_product(
    fn: Callable[[Any, Any, Any], Any], expr: ClauseElement
) -> Iterator[ClauseElement]: ...
def find_tables(
    clause: Union[ClauseElement, FromClause],
    check_columns: bool = ...,
    include_aliases: bool = ...,
    include_joins: bool = ...,
    include_selects: bool = ...,
    include_crud: bool = ...,
) -> List[Table]: ...
def unwrap_order_by(clause: ClauseElement) -> List[ClauseElement]: ...
def unwrap_label_reference(element: _T) -> _T: ...
def expand_column_list_from_order_by(
    collist: ColumnCollection[Any], order_by: ClauseElement
) -> List[ClauseElement]: ...
def clause_is_present(clause: Selectable, search: Selectable) -> bool: ...
def tables_from_leftmost(clause: Table) -> Iterator[Table]: ...
def surface_selectables(clause: Any) -> Iterator[FromClause]: ...
def surface_selectables_only(
    clause: Any,
) -> Iterator[Union[Alias, FromClause, ColumnClause[Any]]]: ...
def extract_first_column_annotation(
    column: Any, annotation_name: str
) -> Optional[Any]: ...
def selectables_overlap(left: Any, right: Any) -> bool: ...
def bind_values(clause: Any) -> List[Any]: ...

class _repr_base:
    def trunc(self, value: Any) -> str: ...

class _repr_row(_repr_base):
    row: Any = ...
    max_chars: int = ...
    def __init__(self, row: Any, max_chars: int = ...) -> None: ...

class _repr_params(_repr_base):
    params: Any = ...
    ismulti: Optional[bool] = ...
    batches: Any = ...
    max_chars: int = ...
    def __init__(
        self,
        params: Any,
        batches: Any,
        max_chars: int = ...,
        ismulti: Optional[bool] = ...,
    ) -> None: ...

def adapt_criterion_to_null(crit: Any, nulls: Any) -> Any: ...
def splice_joins(
    left: Any, right: Any, stop_on: Optional[Any] = ...
) -> Any: ...
def reduce_columns(
    columns: Any, *clauses: Any, **kw: Any
) -> ColumnSet[Any]: ...
def criterion_as_pairs(
    expression: Any,
    consider_as_foreign_keys: Optional[Any] = ...,
    consider_as_referenced_keys: Optional[Any] = ...,
    any_operator: bool = ...,
) -> Any: ...

class ClauseAdapter(visitors.ReplacingExternalTraversal):
    __traverse_options__: Any = ...
    selectable: Any = ...
    include_fn: Any = ...
    exclude_fn: Any = ...
    equivalents: Any = ...
    adapt_on_names: Any = ...
    def __init__(
        self,
        selectable: Any,
        equivalents: Optional[Any] = ...,
        include_fn: Optional[Any] = ...,
        exclude_fn: Optional[Any] = ...,
        adapt_on_names: bool = ...,
        anonymize_labels: bool = ...,
    ) -> None: ...
    def replace(self, col: Any) -> Optional[Any]: ...

class ColumnAdapter(ClauseAdapter):
    columns: Any = ...
    adapt_required: Any = ...
    allow_label_resolve: Any = ...
    def __init__(
        self,
        selectable: Any,
        equivalents: Optional[Any] = ...,
        adapt_required: bool = ...,
        include_fn: Optional[Any] = ...,
        exclude_fn: Optional[Any] = ...,
        adapt_on_names: bool = ...,
        allow_label_resolve: bool = ...,
        anonymize_labels: bool = ...,
    ) -> None: ...
    class _IncludeExcludeMapping:
        parent: Any = ...
        columns: Any = ...
        def __init__(self, parent: Any, columns: Any) -> None: ...
        def __getitem__(self, key: Any) -> Any: ...
    def wrap(self: _COLA, adapter: Any) -> _COLA: ...
    def traverse(self, obj: Any) -> Any: ...
    def adapt_clause(self, obj: Any) -> Any: ...
    adapt_list = ClauseAdapter.copy_and_process
    def adapt_check_present(self, col: Any) -> Optional[Any]: ...



from typing import Any

from .base import MSDialect as MSDialect
from .base import MSIdentifierPreparer as MSIdentifierPreparer
from ... import processors as processors
from ... import types as sqltypes
from ... import util as util

class _MSNumeric_pymssql(sqltypes.Numeric):
    def result_processor(self, dialect: Any, type_: Any): ...

class MSIdentifierPreparer_pymssql(MSIdentifierPreparer):
    def __init__(self, dialect: Any) -> None: ...

class MSDialect_pymssql(MSDialect):
    supports_native_decimal: bool = ...
    driver: str = ...
    preparer: Any = ...
    colspecs: Any = ...
    @classmethod
    def dbapi(cls): ...
    def create_connect_args(self, url: Any): ...
    def is_disconnect(self, e: Any, connection: Any, cursor: Any): ...
    def set_isolation_level(self, connection: Any, level: Any) -> None: ...

dialect = MSDialect_pymssql



from typing import AbstractSet
from typing import Any
from typing import Callable
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

from typing_extensions import Literal

from . import _BackrefResult
from . import attributes as attributes
from .base import state_str as state_str
from .interfaces import MANYTOMANY as MANYTOMANY
from .interfaces import MANYTOONE as MANYTOONE
from .interfaces import ONETOMANY as ONETOMANY
from .interfaces import PropComparator as PropComparator
from .interfaces import StrategizedProperty as StrategizedProperty
from .mapper import Mapper
from .util import AliasedInsp as AliasedInsp
from .util import CascadeOptions as CascadeOptions
from .. import log as log
from .. import schema as schema
from .. import sql as sql
from .. import util as util
from ..inspection import inspect as inspect
from ..schema import Column
from ..sql import coercions as coercions
from ..sql import expression as expression
from ..sql import operators as operators
from ..sql import roles as roles
from ..sql import visitors as visitors
from ..sql.util import adapt_criterion_to_null as adapt_criterion_to_null
from ..sql.util import ClauseAdapter as ClauseAdapter
from ..sql.util import join_condition as join_condition
from ..sql.util import selectables_overlap as selectables_overlap
from ..sql.util import visit_binary_product as visit_binary_product

def remote(expr: Any): ...
def foreign(expr: Any): ...

_T = TypeVar("_T")

_InfoDict = Mapping[Any, Any]

class RelationshipProperty(StrategizedProperty[_T]):
    strategy_wildcard_key: str
    inherit_cache: bool
    uselist: Optional[bool]
    argument: Any
    secondary: Any
    primaryjoin: Any
    secondaryjoin: Any
    post_update: bool
    direction: Any
    viewonly: Any
    sync_backref: Any
    lazy: str
    single_parent: Any
    collection_class: Optional[Union[Type[Any], Callable[[], Any]]]
    passive_deletes: Union[bool, Literal["all"]]
    cascade_backrefs: bool
    passive_updates: bool
    remote_side: Any
    enable_typechecks: bool  # NOTE: not documented
    query_class: Any
    innerjoin: Union[bool, str]
    distinct_target_key: Optional[bool]
    doc: Optional[str]
    active_history: bool
    join_depth: Optional[int]
    omit_join: Optional[Literal[False]]
    local_remote_pairs: Any  # NOTE: not documented
    bake_queries: bool
    load_on_pending: bool
    comparator_factory: Any
    comparator: Any
    info: _InfoDict  # NOTE: not set if constructor argument is ``None``
    strategy_key: Tuple[Tuple[str, str]]  # NOTE: not documented
    order_by: Any
    back_populates: Union[None, str]
    backref: Union[None, str, _BackrefResult]
    def __init__(
        self,
        argument: Any,
        secondary: Optional[Any] = ...,
        primaryjoin: Optional[Any] = ...,
        secondaryjoin: Optional[Any] = ...,
        foreign_keys: Optional[Any] = ...,
        uselist: Optional[bool] = ...,
        order_by: Union[
            Literal[False], str, Column, Callable[[], Column]
        ] = ...,
        backref: Union[str, _BackrefResult] = ...,
        back_populates: str = ...,
        overlaps: Union[AbstractSet[str], str] = ...,
        post_update: bool = ...,
        cascade: Union[Literal[False], Sequence[str]] = ...,
        viewonly: bool = ...,
        lazy: str = ...,
        collection_class: Optional[Union[Type[Any], Callable[[], Any]]] = ...,
        passive_deletes: Union[bool, Literal["all"]] = ...,
        passive_updates: bool = ...,
        remote_side: Optional[Any] = ...,
        enable_typechecks: bool = ...,  # NOTE: not documented
        join_depth: Optional[int] = ...,
        comparator_factory: Optional[Any] = ...,
        single_parent: bool = ...,
        innerjoin: Union[bool, str] = ...,
        distinct_target_key: Optional[bool] = ...,
        doc: Optional[str] = ...,
        active_history: bool = ...,
        cascade_backrefs: bool = ...,
        load_on_pending: bool = ...,
        bake_queries: bool = ...,
        _local_remote_pairs: Optional[Any] = ...,
        query_class: Optional[Any] = ...,
        info: Optional[_InfoDict] = ...,
        omit_join: Optional[Literal[False]] = ...,
        sync_backref: Optional[Any] = ...,
    ) -> None: ...
    def instrument_class(self, mapper: Any) -> None: ...
    class Comparator(PropComparator):
        prop: Any = ...
        def __init__(
            self,
            prop: Any,
            parentmapper: Any,
            adapt_to_entity: Optional[Any] = ...,
            of_type: Optional[Any] = ...,
            extra_criteria: Any = ...,
        ) -> None: ...
        def adapt_to_entity(self, adapt_to_entity: Any): ...
        def entity(self): ...
        def mapper(self): ...
        def __clause_element__(self): ...
        def of_type(self, cls: Any): ...
        def and_(self, *other: Any): ...
        def in_(self, other: Any) -> None: ...
        __hash__: Any = ...
        def __eq__(self, other: Any) -> Any: ...
        def any(self, criterion: Optional[Any] = ..., **kwargs: Any): ...
        def has(self, criterion: Optional[Any] = ..., **kwargs: Any): ...
        def contains(self, other: Any, **kwargs: Any): ...
        def __ne__(self, other: Any) -> Any: ...
        def property(self): ...
    def merge(
        self,
        session: Any,
        source_state: Any,
        source_dict: Any,
        dest_state: Any,
        dest_dict: Any,
        load: Any,
        _recursive: Any,
        _resolve_conflict_map: Any,
    ) -> None: ...
    def cascade_iterator(
        self,
        type_: Any,
        state: Any,
        dict_: Any,
        visited_states: Any,
        halt_on: Optional[Any] = ...,
    ) -> None: ...
    def entity(self) -> Union[AliasedInsp, Mapper]: ...
    def mapper(self) -> Mapper: ...
    def do_init(self) -> None: ...
    @property
    def cascade(self) -> CascadeOptions: ...
    @cascade.setter
    def cascade(self, cascade: Sequence[str]) -> None: ...

class JoinCondition:
    parent_persist_selectable: Any = ...
    parent_local_selectable: Any = ...
    child_persist_selectable: Any = ...
    child_local_selectable: Any = ...
    parent_equivalents: Any = ...
    child_equivalents: Any = ...
    primaryjoin: Any = ...
    secondaryjoin: Any = ...
    secondary: Any = ...
    consider_as_foreign_keys: Any = ...
    prop: Any = ...
    self_referential: Any = ...
    support_sync: Any = ...
    can_be_synced_fn: Any = ...
    def __init__(
        self,
        parent_persist_selectable: Any,
        child_persist_selectable: Any,
        parent_local_selectable: Any,
        child_local_selectable: Any,
        primaryjoin: Optional[Any] = ...,
        secondary: Optional[Any] = ...,
        secondaryjoin: Optional[Any] = ...,
        parent_equivalents: Optional[Any] = ...,
        child_equivalents: Optional[Any] = ...,
        consider_as_foreign_keys: Optional[Any] = ...,
        local_remote_pairs: Optional[Any] = ...,
        remote_side: Optional[Any] = ...,
        self_referential: bool = ...,
        prop: Optional[Any] = ...,
        support_sync: bool = ...,
        can_be_synced_fn: Any = ...,
    ): ...
    @property
    def primaryjoin_minus_local(self): ...
    @property
    def secondaryjoin_minus_local(self): ...
    def primaryjoin_reverse_remote(self): ...
    def remote_columns(self): ...
    def local_columns(self): ...
    def foreign_key_columns(self): ...
    def join_targets(
        self,
        source_selectable: Any,
        dest_selectable: Any,
        aliased: Any,
        single_crit: Optional[Any] = ...,
        extra_criteria: Any = ...,
    ): ...
    def create_lazy_clause(self, reverse_direction: bool = ...): ...

class _ColInAnnotations:
    name: Any = ...
    def __init__(self, name: Any) -> None: ...
    def __call__(self, c: Any): ...



from typing import Any
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import TypeVar

from ...engine.result import FilterResult
from ...engine.result import FrozenResult
from ...engine.result import MergedResult
from ...engine.result import Result
from ...engine.result import RMKeyView
from ...engine.result import Row

_TAsyncResult = TypeVar("_TAsyncResult", bound=AsyncResult)
_TAsyncScalarResult = TypeVar("_TAsyncScalarResult", bound=AsyncScalarResult)
_TAsyncMappingResult = TypeVar(
    "_TAsyncMappingResult", bound=AsyncMappingResult
)

class AsyncCommon(FilterResult):
    async def close(self) -> None: ...

class AsyncResult(AsyncCommon):
    def __init__(self, real_result: Result) -> None: ...
    def keys(self) -> RMKeyView: ...
    def unique(
        self: _TAsyncResult, strategy: Optional[Any] = ...
    ) -> _TAsyncResult: ...
    def columns(
        self: _TAsyncResult, *col_expressions: Any
    ) -> _TAsyncResult: ...
    async def partitions(
        self, size: Optional[int] = ...
    ) -> Iterator[List[Any]]: ...
    async def fetchone(self) -> Optional[Row]: ...
    async def fetchmany(self, size: Optional[int] = ...) -> List[Row]: ...
    async def all(self) -> List[Row]: ...
    def __aiter__(self: _TAsyncResult) -> _TAsyncResult: ...
    async def __anext__(self) -> Row: ...
    async def first(self) -> Optional[Row]: ...
    async def one_or_none(self) -> Optional[Row]: ...
    async def scalar_one(self) -> Any: ...
    async def scalar_one_or_none(self) -> Optional[Any]: ...
    async def one(self) -> Row: ...
    async def scalar(self) -> Optional[Any]: ...
    async def freeze(self) -> FrozenResult: ...
    def merge(self, *others: Any) -> MergedResult: ...
    def scalars(self, index: int = ...) -> AsyncScalarResult: ...
    def mappings(self) -> AsyncMappingResult: ...

class AsyncScalarResult(AsyncCommon):
    def __init__(self, real_result: AsyncResult, index: Any) -> None: ...
    def unique(
        self: _TAsyncScalarResult, strategy: Optional[Any] = ...
    ) -> _TAsyncScalarResult: ...
    async def partitions(
        self, size: Optional[int] = ...
    ) -> Iterator[List[Any]]: ...
    async def fetchall(self) -> List[Any]: ...
    async def fetchmany(self, size: Optional[int] = ...) -> List[Any]: ...
    async def all(self) -> List[Any]: ...
    def __aiter__(self: _TAsyncScalarResult) -> _TAsyncScalarResult: ...
    async def __anext__(self) -> Any: ...
    async def first(self) -> Optional[Any]: ...
    async def one_or_none(self) -> Optional[Any]: ...
    async def one(self) -> Any: ...

class AsyncMappingResult(AsyncCommon):
    def __init__(self, result: AsyncResult) -> None: ...
    def keys(self) -> RMKeyView: ...
    def unique(
        self: _TAsyncMappingResult, strategy: Optional[Any] = ...
    ) -> _TAsyncMappingResult: ...
    def columns(
        self: _TAsyncMappingResult, *col_expressions: Any
    ) -> _TAsyncMappingResult: ...
    async def partitions(
        self, size: Optional[int] = ...
    ) -> Iterator[List[Mapping[Any, Any]]]: ...
    async def fetchall(self) -> List[Mapping[Any, Any]]: ...
    async def fetchone(self) -> Optional[Mapping[Any, Any]]: ...
    async def fetchmany(
        self, size: Optional[int] = ...
    ) -> List[Mapping[Any, Any]]: ...
    async def all(self) -> List[Mapping[Any, Any]]: ...
    def __aiter__(self: _TAsyncMappingResult) -> _TAsyncMappingResult: ...
    async def __anext__(self) -> Mapping[Any, Any]: ...
    async def first(self) -> Optional[Mapping[Any, Any]]: ...
    async def one_or_none(self) -> Optional[Mapping[Any, Any]]: ...
    async def one(self) -> Mapping[Any, Any]: ...
